from __future__ import annotations

import json
from dataclasses import dataclass, field
from typing import Annotated, Any, Dict, Tuple, Type

import numpy as np
import pandas as pd
import xarray as xa

from openlifu.plan.param_constraint import PARAM_STATUS_SYMBOLS, ParameterConstraint
from openlifu.util.annotations import OpenLIFUFieldData
from openlifu.util.dict_conversion import DictMixin
from openlifu.util.units import getunitconversion, getunittype

DEFAULT_ORIGIN = np.zeros(3)

PARAM_FORMATS = {
    "mainlobe_pnp_MPa": ["max", "0.3f", "MPa", "Mainlobe Peak Negative Pressure"],
    "mainlobe_isppa_Wcm2": ["max", "0.1f", "W/cm^2", "Mainlobe I_SPPA"],
    "mainlobe_ispta_mWcm2": ["mean", "0.1f", "mW/cm^2", "Mainlobe I_SPTA"],
    "target_position_lat_mm": ["mean", "0.1f", "mm", "Target Position (Lateral)"],
    "target_position_ele_mm": ["mean", "0.1f", "mm", "Target Position (Elevation)"],
    "target_position_ax_mm": ["mean", "0.1f", "mm", "Target Position (Axial)"],
    "focal_centroid_lat_mm": ["mean", "0.1f", "mm", "Focal Centroid (Lateral)"],
    "focal_centroid_ele_mm": ["mean", "0.1f", "mm", "Focal Centroid (Elevation)"],
    "focal_centroid_ax_mm": ["mean", "0.1f", "mm", "Focal Centroid (Axial)"],
    "beamwidth_lat_3dB_mm": ["mean", "0.2f", "mm", "3dB Beamwidth (Lateral)"],
    "beamwidth_ele_3dB_mm": ["mean", "0.2f", "mm", "3dB Beamwidth (Elevational)"],
    "beamwidth_ax_3dB_mm": ["mean", "0.2f", "mm", "3dB Beamwidth (Axial)"],
    "beamwidth_lat_6dB_mm": ["mean", "0.2f", "mm", "6dB Beamwidth (Lateral)"],
    "beamwidth_ele_6dB_mm": ["mean", "0.2f", "mm", "6dB Beamwidth (Elevational)"],
    "beamwidth_ax_6dB_mm": ["mean", "0.2f", "mm", "6dB Beamwidth (Axial)"],
    "sidelobe_pnp_MPa": ["max", "0.3f", "MPa", "Sidelobe Peak Negative Pressure"],
    "sidelobe_isppa_Wcm2": ["max", "0.1f", "W/cm^2", "Sidelobe I_SPPA"],
    "sidelobe_to_mainlobe_pressure_ratio": ["mean", "0.2f", "", "Sidelobe/Mainlobe Pressure Ratio"],
    "sidelobe_to_mainlobe_intensity_ratio": ["mean", "0.2f", "", "Sidelobe/Mainlobe Intensity Ratio"],
    "global_pnp_MPa": ["max", "0.3f", "MPa", "Global Peak Negative Pressure"],
    "global_isppa_Wcm2": ["max", "0.1f", "W/cm^2", "Global I_SPPA"],
    "global_ispta_mWcm2": [None, "0.1f", "mW/cm^2", "Global I_SPTA"],
    "MI": [None, "0.2f", "", "MI"],
    "TIC": [None, "0.2f", "", "TIC"],
    "voltage_V": [None, "0.1f", "V", "Voltage"],
    "p0_MPa": ["max", "0.3f", "MPa", "Emitted Pressure"],
    "power_W": [None, "0.2f", "W", "Emitted Power"],
    "duty_cycle_pulse_train_pct": [None, "0.1f", "%", "Pulse Train Duty Cycle"],
    "duty_cycle_sequence_pct": [None, "0.1f", "%", "Sequence Duty Cycle"],
    "sequence_duration_s": [None, "0.0f", "s", "Sequence Duration"],}

@dataclass
class SolutionAnalysis(DictMixin):
    mainlobe_pnp_MPa: Annotated[list[float], OpenLIFUFieldData("Mainlobe PNP", "Peak negative pressure in the mainlobe, in MPa")] = field(default_factory=list)
    """Peak negative pressure in the mainlobe, in MPa"""

    mainlobe_isppa_Wcm2: Annotated[list[float], OpenLIFUFieldData("Mainlobe ISPPA", "Spatial peak pulse average intensity in the mainlobe, in W/cm²")] = field(default_factory=list)
    """Spatial peak pulse average intensity in the mainlobe, in W/cm²"""

    mainlobe_ispta_mWcm2: Annotated[list[float], OpenLIFUFieldData("Mainlobe ISPTA", "Spatial peak time average intensity in the mainlobe, in mW/cm²")] = field(default_factory=list)
    """Spatial peak time average intensity in the mainlobe, in mW/cm²"""

    target_position_lat_mm: Annotated[list[float], OpenLIFUFieldData("Target position lateral (mm)", "Lateral position of the target, in mm")] = field(default_factory=list)
    """Lateral position of the target, in mm"""

    target_position_ele_mm: Annotated[list[float], OpenLIFUFieldData("Target position elevation (mm)", "Elevation position of the target, in mm")] = field(default_factory=list)
    """Elevation position of the target, in mm"""

    target_position_ax_mm: Annotated[list[float], OpenLIFUFieldData("Target position axial (mm)", "Axial position of the target, in mm")] = field(default_factory=list)
    """Axial position of the target, in mm"""

    focal_centroid_lat_mm: Annotated[list[float], OpenLIFUFieldData("Focal centroid lateral (mm)", "Lateral centroid of the focus, in mm")] = field(default_factory=list)
    """Lateral centroid of the focus, in mm"""

    focal_centroid_ele_mm: Annotated[list[float], OpenLIFUFieldData("Focal centroid elevation (mm)", "Elevation centroid of the focus, in mm")] = field(default_factory=list)
    """Elevation centroid of the focus, in mm"""

    focal_centroid_ax_mm: Annotated[list[float], OpenLIFUFieldData("Focal centroid axial (mm)", "Axial centroid of the focus, in mm")] = field(default_factory=list)
    """Axial centroid of the focus, in mm"""

    beamwidth_lat_3dB_mm: Annotated[list[float], OpenLIFUFieldData("3dB lateral beamwidth", "Lateral beamwidth at -3 dB, in mm")] = field(default_factory=list)
    """Lateral beamwidth at -3 dB, in mm"""

    beamwidth_ele_3dB_mm: Annotated[list[float], OpenLIFUFieldData("3dB elevation beamwidth", "Elevation beamwidth at -3 dB, in mm")] = field(default_factory=list)
    """Elevation beamwidth at -3 dB, in mm"""

    beamwidth_ax_3dB_mm: Annotated[list[float], OpenLIFUFieldData("3dB axial beamwidth", "Axial beamwidth at -3 dB, in mm")] = field(default_factory=list)
    """Axial beamwidth at -3 dB, in mm"""

    beamwidth_lat_6dB_mm: Annotated[list[float], OpenLIFUFieldData("6dB lateral beamwidth", "Lateral beamwidth at -6 dB, in mm")] = field(default_factory=list)
    """Lateral beamwidth at -6 dB, in mm"""

    beamwidth_ele_6dB_mm: Annotated[list[float], OpenLIFUFieldData("6dB elevation beamwidth", "Elevation beamwidth at -6 dB, in mm")] = field(default_factory=list)
    """Elevation beamwidth at -6 dB, in mm"""

    beamwidth_ax_6dB_mm: Annotated[list[float], OpenLIFUFieldData("6dB axial beamwidth", "Axial beamwidth at -6 dB, in mm")] = field(default_factory=list)
    """Axial beamwidth at -6 dB, in mm"""

    sidelobe_pnp_MPa: Annotated[list[float], OpenLIFUFieldData("Sidelobe PNP", "Peak negative pressure in the sidelobes, in MPa")] = field(default_factory=list)
    """Peak negative pressure in the sidelobes, in MPa"""

    sidelobe_isppa_Wcm2: Annotated[list[float], OpenLIFUFieldData("Sidelobe ISPPA", "Spatial peak pulse average intensity in the sidelobes, in W/cm²")] = field(default_factory=list)
    """Spatial peak pulse average intensity in the sidelobes, in W/cm²"""

    sidelobe_to_mainlobe_pressure_ratio: Annotated[list[float], OpenLIFUFieldData("Sidelobe/Mainlobe Pressure Ratio", "Ratio of sidelobe peak negative pressure to mainlobe peak negative pressure")] = field(default_factory=list)
    """Ratio of sidelobe peak negative pressure to mainlobe peak negative pressure"""

    sidelobe_to_mainlobe_intensity_ratio: Annotated[list[float], OpenLIFUFieldData("Sidelobe/Mainlobe Intensity Ratio", "Ratio of sidelobe spatial peak pulse average intensity to mainlobe spatial peak pulse average intensity")] = field(default_factory=list)
    """Ratio of sidelobe spatial peak pulse average intensity to mainlobe spatial peak pulse average intensity"""

    global_pnp_MPa: Annotated[list[float], OpenLIFUFieldData("Global PNP", "Maximum peak negative pressure in the entire field, in MPa")] = field(default_factory=list)
    """Maximum peak negative pressure in the entire field, in MPa"""

    global_isppa_Wcm2: Annotated[list[float], OpenLIFUFieldData("Global ISPPA", "Maximum spatial peak pulse average intensity in the entire field, in W/cm²")] = field(default_factory=list)
    """Maximum spatial peak pulse average intensity in the entire field, in W/cm²"""

    global_ispta_mWcm2: Annotated[float | None, OpenLIFUFieldData("Global ISPTA (mW/cm²)", "Global Intensity at Spatial-Peak, Time-Average (I_SPTA) (mW/cm²)")] = None
    """Global Intensity at Spatial-Peak, Time-Average (I_SPTA) (mW/cm²)"""

    MI: Annotated[float | None, OpenLIFUFieldData("Mechanical index (MI)", "Mechanical index (MI)")] = None
    """Mechanical index (MI)"""

    TIC: Annotated[float | None, OpenLIFUFieldData("Thermal index (TIC)", "Thermal index in cranium (TIC)")] = None
    """Thermal index in cranium (TIC)"""

    voltage_V: Annotated[float | None, OpenLIFUFieldData("Voltage (V)", "Voltage applied to the transducer (V)")] = None
    """Voltage applied to the transducer (V)"""

    p0_MPa: Annotated[list[float], OpenLIFUFieldData("Emitted pressure (MPa)", "Initial pressure values in the field, in MPa")] = field(default_factory=list)
    """Initial pressure values in the field (MPa)"""

    power_W: Annotated[float | None, OpenLIFUFieldData("Emitted Power (W)", "Emitted power from the transducer face (W)")] = None
    """Emitted power from the transducer face (W)"""

    duty_cycle_pulse_train_pct: Annotated[float | None, OpenLIFUFieldData("Duty cycle pulse train", "Duty cycle within a pulse train (0-1)")] = None
    """Duty cycle within a pulse train (0-1)"""

    duty_cycle_sequence_pct: Annotated[float | None, OpenLIFUFieldData("Duty cycle sequence", "Duty cycle of the overall sequence (0-1)")] = None
    """Duty cycle of the overall sequence (0-1)"""

    sequence_duration_s: Annotated[float | None, OpenLIFUFieldData("Sequence duration (s)", "Total duration of the sequence (s)")] = None
    """Total duration of the sequence (s)"""


    param_constraints: Annotated[Dict[str, ParameterConstraint], OpenLIFUFieldData("Parameter constraints", None)] = field(default_factory=dict)
    """TODO: Add description"""

    def to_table(self, constraints:Dict[str,ParameterConstraint]|None=None, focus_index=None) -> pd.DataFrame:
        records = []
        if constraints is None:
            constraints = self.param_constraints
        for p in constraints:
            if p not in PARAM_FORMATS:
                raise ValueError(f"Unknown parameter constraint for '{p}'. Must be one of: {list(PARAM_FORMATS.keys())}")
        for param, fmt in PARAM_FORMATS.items():
            value_by_focus = None
            agg_value = None
            if fmt[0] is None:
                value_by_focus = None
                agg_value = self.__dict__[param]
            elif fmt[0] == "max":
                value_by_focus = self.__dict__[param]
                agg_value = max(value_by_focus)
            elif fmt[0] == "mean":
                value_by_focus = self.__dict__[param]
                agg_value = np.mean(value_by_focus)
            else:
                raise ValueError(f"Unknown aggregation method '{fmt[0]}' for parameter '{param}'.")
            if agg_value is not None:
                record = {"id": param,
                          "Param": fmt[3],
                          "Value" : "",
                          "Units" : fmt[2],
                          "Status": "",
                          "_value" : agg_value,
                          "_value_by_focus": value_by_focus,
                          "_warning": False,
                          "_error": False}
                if np.isnan(agg_value):
                    record["Value"] = "NaN"
                else:
                    if focus_index is None:
                        record["Value"] = f"{agg_value:{fmt[1]}}"
                    elif value_by_focus is None:
                        record["Value"] = "N/A"
                    else:
                        value = value_by_focus[focus_index]
                        record["Value"] = f"{value:{fmt[1]}}"
                    if param in constraints:
                        record['_warning'] = constraints[param].is_warning(agg_value)
                        record['_error'] = constraints[param].is_error(agg_value)
                        record["Status"] = PARAM_STATUS_SYMBOLS[constraints[param].get_status(agg_value)]
                records.append(record)
        return pd.DataFrame.from_records(records)

    @classmethod
    def from_dict(cls: Type[SolutionAnalysis], parameter_dict: Dict[str, Any]) -> SolutionAnalysis:
        """
        Create an object from a dictionary

        Args:
            parameter_dict: dictionary of parameters to define the object
        Returns: new object
        """
        parameter_dict["param_constraints"] = {
            k: ParameterConstraint.from_dict(v)
            for k, v in parameter_dict.get("param_constraints", {}).items()
        }

        return cls(**parameter_dict)

    @staticmethod
    def from_json(json_string : str) -> SolutionAnalysis:
        """Load a SolutionAnalysis from a json string"""
        return SolutionAnalysis.from_dict(json.loads(json_string))

    def to_json(self, compact:bool) -> str:
        """Serialize a SolutionAnalysis to a json string

        Args:
            compact: if enabled then the string is compact (not pretty). Disable for pretty.

        Returns: A json string representing the complete SolutionAnalysis object.
        """
        if compact:
            return json.dumps(self.to_dict(), separators=(',', ':'))
        else:
            return json.dumps(self.to_dict(), indent=4)

@dataclass
class SolutionAnalysisOptions(DictMixin):
    standoff_sound_speed: Annotated[float, OpenLIFUFieldData("Standoff sound speed (m/s)", "Speed of sound in standoff, for calculating initial impedance")] = 1500.0
    """Speed of sound in standoff, for calculating initial impedance"""

    standoff_density: Annotated[float, OpenLIFUFieldData("Standoff density (kg/m³)", "Density of standoff medium (kg/m³)")] = 1000.0
    """Density of standoff medium (kg/m³)"""

    ref_sound_speed: Annotated[float, OpenLIFUFieldData("Reference sound speed (m/s)", "Reference speed of sound in the medium (m/s)")] = 1500.0
    """Reference speed of sound in the medium (m/s)"""

    ref_density: Annotated[float, OpenLIFUFieldData("Reference density (kg/m³)", "Reference density (kg/m³)")] = 1000.0
    """Reference density (kg/m³)"""

    mainlobe_aspect_ratio: Annotated[Tuple[float, float, float], OpenLIFUFieldData("Mainlobe aspect ratio (lat,ele,ax)", "Aspect ratio of the mainlobe mask")] = (1., 1., 5.)
    """Aspect ratio of the mainlobe ellipsoid mask, in the form (lat,ele,ax). (1,1,5) means an ellipsoid 5x as long as it is wide."""

    mainlobe_radius: Annotated[float, OpenLIFUFieldData("Mainlobe mask radius", "Size of the mainlobe mask, in the units provided for Distance units (`distance_units`)")] = 2.5e-3
    """Size of the mainlobe mask, in the units provided for Distance units (`distance_units`). The mainlobe mask is an ellipsoid with this radius, scaled by the `mainlobe_aspect_ratio`."""

    beamwidth_radius: Annotated[float, OpenLIFUFieldData("Beamwidth search radius", "Size of the beamwidth search, in the units provided for Distance units (`distance_units`)")] = 5e-3
    """Size of the beamwidth search, in the units provided for Distance units (`distance_units`). The beamwidth is found along the lateral and elevation lines perpendicular to the focus axis."""

    sidelobe_radius: Annotated[float, OpenLIFUFieldData("Sidelobe radius", "Size of the sidelobe mask, in the units provided for Distance units (`distance_units`)")] = 3e-3
    """Size of the sidelobe mask, in the units provided for Distance units (`distance_units`). Pressure outside of this ellipsoid (scaled by `mainlobe_aspect_ratio`) is considered outside of the focal region."""

    sidelobe_zmin: Annotated[float, OpenLIFUFieldData("Sidelobe minimum z", "Minimum z coordinate of the sidelobe mask, in the units provided for Distance units (`distance_units`)")] = 1e-3
    """Minimum z coordinate of the sidelobe mask, in the units provided for Distance units (`distance_units`). This value is used to ignore emitted pressure artifacts."""

    distance_units: Annotated[str, OpenLIFUFieldData("Distance units", "The units used for distance measurements")] = "m"
    """The units used for distance measurements"""

    param_constraints: Annotated[Dict[str, ParameterConstraint], OpenLIFUFieldData("Parameter constraints", None)] = field(default_factory=dict)
    """TODO: Add description"""

    def __post_init__(self):
        if self.standoff_sound_speed <= 0:
            raise ValueError("Standoff sound speed must be greater than 0")
        if self.standoff_density <= 0:
            raise ValueError("Standoff density must be greater than 0")
        if self.ref_sound_speed <= 0:
            raise ValueError("Reference sound speed must be greater than 0")
        if self.ref_density <= 0:
            raise ValueError("Reference density must be greater than 0")
        if not isinstance(self.mainlobe_aspect_ratio, tuple | list) or len(self.mainlobe_aspect_ratio) != 3:
            raise TypeError("Mainlobe aspect ratio must be a tuple or list of three floats (lat, ele, ax)")
        self.mainlobe_aspect_ratio = tuple(self.mainlobe_aspect_ratio)  # Ensure it's a tuple
        if not all(isinstance(x, int | float) for x in self.mainlobe_aspect_ratio):
            raise TypeError("Mainlobe aspect ratio must contain only numbers")
        if not isinstance(self.mainlobe_radius, int | float) or self.mainlobe_radius <= 0:
            raise ValueError("Mainlobe radius must be a positive number")
        if not isinstance(self.beamwidth_radius, int | float) or self.beamwidth_radius <= 0:
            raise ValueError("Beamwidth radius must be a positive number")
        if not isinstance(self.sidelobe_radius, int | float) or self.sidelobe_radius <= 0:
            raise ValueError("Sidelobe radius must be a positive number")
        if not isinstance(self.sidelobe_zmin, int | float) or self.sidelobe_zmin < 0:
            raise ValueError("Sidelobe minimum z must be a non-negative number")
        if not isinstance(self.distance_units, str):
            raise TypeError("Distance units must be a string")
        if getunittype(self.distance_units) != 'distance':
            raise ValueError(f"Distance units must be a length unit, got {self.distance_units}")

    @classmethod
    def from_dict(cls: Type[SolutionAnalysisOptions], parameter_dict: Dict[str, Any]) -> SolutionAnalysisOptions:
        """
        Create an object from a dictionary

        Args:
            parameter_dict: dictionary of parameters to define the object
        Returns: new object
        """
        parameter_dict["param_constraints"] = {
            k: ParameterConstraint.from_dict(v)
            for k, v in parameter_dict.get("param_constraints", {}).items()
        }

        return cls(**parameter_dict)

def find_centroid(da: xa.DataArray, cutoff:float, units:None) -> np.ndarray:
    """Find the centroid of a thresholded region of a DataArray"""
    if units is not None and getunittype(units) != 'distance':
        raise ValueError(f"Units must be a length unit, got {units}")
    da = da.where(da > cutoff, 0)
    dims = list(da.dims)
    coords = np.meshgrid(*[da.coords[coord] for coord in dims], indexing='ij')
    centroid = np.array([np.sum(da*coords[dims.index(dim)])/da.sum() for dim in da.dims])
    if units is not None:
        da_units = [da.coords[dim].attrs.get('units', None) for dim in dims]
        centroid = np.array([getunitconversion(coord_units, units) * c for coord_units, c in zip(da_units, centroid)])
    return centroid

def get_focus_matrix(focus, origin=[0,0,0]) -> np.ndarray:
    """Get the coordinate transform from the focus to the transducer

    The "focal coordinate system" here refers to a coordinate system whose z-axis is along
    the ray from the transducer's "effective origin" (see `Transducer.get_effective_origin`)
    to the focus center.

    Args:
        focus: A 3D point describing the focal coordinate system location in transducer coordinates
        origin: A 3D point describing the transducer "effective origin" in transducer coordinates

    Returns: A 4x4 affine transform matrix that describes the coordinate transformation from the focus
        coordinate system to the transducer coordinates.
    """
    focus = np.array(focus)
    origin = np.array(origin)
    zvec = (focus-origin)/np.linalg.norm(focus-origin)
    az = -np.arctan2(zvec[0],zvec[2])
    xvec = np.array([np.cos(az), 0, np.sin(az)])
    yvec = np.cross(zvec, xvec)
    uv = np.array([xvec, yvec, zvec, focus])
    M = np.concatenate([uv.T,np.zeros([1,4])],axis=0)
    M[3,3] = 1
    return M

def get_gridded_transformed_coords(da: xa.DataArray, matrix: np.ndarray, as_dataset=True):
    """Transform the coords of a DataArray using a transform matrix.

    Args:
        da: DataArray whose coordinates will be used
        matrix: a 4x4 coordinate transformation matrix, transforming from the desired coordinate system
            to the coordinate system of `da`
        as_dataset: Whether to return the transformed coords as a numpy array or an xarray Dataset

    Returns the transformed coordinate grid as either a numpy array or an xarray Dataset.
    """
    shape = tuple(da.sizes[d] for d in da.dims)
    coords = np.meshgrid(*[da.coords[coord] for coord in da.dims], indexing='ij')
    coords = np.concatenate([coord.reshape(-1,1) for coord in coords], axis=1)
    coords = np.concatenate([coords, np.ones((coords.shape[0],1))], axis=1)
    coords = np.dot(np.linalg.inv(matrix), coords.T).T
    coords = coords[:,0:3].reshape(*shape, 3)
    if as_dataset:
        coords = xa.Dataset({f'd_{dim}': (da.dims, coords[...,i]) for i, dim in enumerate(da.dims)}, coords=da.coords)
    return coords

def get_offset_grid(da: xa.DataArray, focus, origin=DEFAULT_ORIGIN, as_dataset=True):
    """Transform the coords of a DataArray that is in transducer coordinates to focus coordinates

    See `get_focus_matrix` for the meaning of "focus coordinates"

    Args:
        da: DataArray whose coordinates will be used (presumably the transducer coordinates)
        focus: A 3D point describing the focus location in the coordinates of `da`
        origin: A 3D point describing the "effective origin" in the coordinates of `da`
            (see `Transducer.get_effective_origin` for the meaning of this).
        as_dataset: Whether to return the transformed coords as a numpy array or an xarray Dataset

    Returns the transformed coordinate grid as either a numpy array or an xarray Dataset.
    """
    M = get_focus_matrix(focus, origin=origin)
    #M = focus.get_matrix()
    coords = get_gridded_transformed_coords(da, M, as_dataset=as_dataset)
    return coords

def calc_dist_from_focus(da: xa.DataArray, focus, origin=DEFAULT_ORIGIN, aspect_ratio=[1,1,1], as_dataarray=True):
    """Compute a distance map from a focus point in transducer space, using a possibly distorted metric that respects
    the symmetry of the focus shape (e.g. it could be cigar-shaped).

    Args:
        da: DataArray that will supply the coordnate grid (presumably transducer coordinates)
        focus: A 3D point describing the focus location in the coordinates of `da`
        origin: A 3D point describing the "effective origin" in the coordinates of `da`
            (see `Transducer.get_effective_origin` for the meaning of this).
        aspect_ratio: x,y,z scalings on the focal coordinate system to distort the space before computing distance
            (see `get_focus_matrix` for the meaning of "focus coordinates").
        as_dataarray: Whether to return the distance map as a numpy array or an xarray DataArray

    Returns the distance map as either a numpy array or an xarray DataArray.
    """
    coords = get_offset_grid(da, focus, origin=origin, as_dataset=False)
    dist = np.sqrt(np.sum((coords/aspect_ratio)**2, axis=-1))
    if as_dataarray:
        dist = xa.DataArray(dist, coords=da.coords, dims=da.dims)
    return dist

def get_mask(
    da: xa.DataArray,
    focus,
    distance:float,
    origin=DEFAULT_ORIGIN,
    aspect_ratio=[1,1,1],
    operator='<',
) -> xa.DataArray:
    """Compute a boolean mask of the focus region in transducer space.

    The focus region is an ellipsoid centered at the focus point.

    Args:
        da: DataArray that will supply the coordnate grid (presumably transducer coordinates)
        focus: A 3D point describing the focus location in the coordinates of `da`
        distance: How far from the `focus` to include points in the mask. See `calc_dist_from_focus`
            for the distorted metric under which a ball of points becomes an ellispoid in euclidean space.
        origin: A 3D point describing the "effective origin" in the coordinates of `da`
            (see `Transducer.get_effective_origin` for the meaning of this).
        aspect_ratio: x,y,z scalings on the focal coordinate system to distort the space before computing distances
            (see `get_focus_matrix` for the meaning of "focus coordinates").
        operator: a string representation of an inequality operator that represents the desired masking operation.
            The default '<' for example includes points strictly *inside* the focal ellipsoid.

    Returns: the focal mask as an xarray DataArray
    """
    dist = calc_dist_from_focus(da, focus, origin=origin, aspect_ratio=aspect_ratio)
    if operator == '<':
        mask = dist < distance
    elif operator == '<=':
        mask = dist <= distance
    elif operator == '>':
        mask = dist > distance
    elif operator == '>=':
        mask = dist >= distance
    else:
        raise ValueError("Operator must be '<', '>', '<=', or '>='.")
    return mask

def interp_transformed_axis(
    da: xa.DataArray,
    focus,
    dim,
    origin=DEFAULT_ORIGIN,
    min_offset:float | None=None,
    max_offset:float | None=None,
) -> xa.DataArray:
    """Interpolate data along a focal coordinate system axis.

    See `get_focus_matrix` for the meaning of "focal coordinate system."

    Args:
        da: DataArray whose values will be sampled (presumably defined on transducer coordinates).
        focus: A 3D point describing the focus location in the coordinates of `da`
        dim: The name of the dimension of `da` whose corresponding focal coordinate system axis should be sampled along.
            For example, the "axial" dimension of a transducer corresponds to the focal axis (z-axis) in the focal coordinate
            system, that is, the ray from the transducer's "effective origin" (see `Transducer.get_effective_origin`)
            to the focus center.
        origin: A 3D point describing the "effective origin" in the coordinates of `da`
            (see `Transducer.get_effective_origin` for the meaning of this).
        min_offset: How far along the negative focal `dim` direction to sample. By default samples as far as the coordinate
            grid allows.
        max_offset: How far along the positive focal `dim` direction to sample. By default samples as far as the coordinate
            grid allows.

    Returns: a 1D DataArray of interpolated values from `da`.
    """
    matrix = get_focus_matrix(focus, origin=origin)
    coords = get_gridded_transformed_coords(da, matrix, as_dataset=True)
    if min_offset is None:
        min_offset = float(coords[f'd_{dim}'].min())
    if max_offset is None:
        max_offset = float(coords[f'd_{dim}'].max())
    n = da.sizes[dim]*2
    interp_dim = np.linspace(min_offset, max_offset, n)
    interp_xyz = np.zeros((n,3))
    interp_xyz[:,da.dims.index(dim)] = interp_dim
    interp_xyz = np.dot(matrix, np.concatenate([interp_xyz.T, np.ones((1,n))], axis=0)).T[:,0:3]
    interpolants = {d: xa.DataArray(interp_xyz[:,i], coords={f'offset_d{dim}':interp_dim}) for i, d in enumerate(da.dims)}
    interp_da = da.interp(**interpolants)
    for d in da.dims:
        interp_da.assign_coords(dim=(f'offset_d{dim}', interpolants[d].to_numpy()))
    return interp_da

def get_beam_bounds(
    da: xa.DataArray,
    focus,
    dim,
    cutoff:float,
    origin=DEFAULT_ORIGIN,
    min_offset:float | None=None,
    max_offset:float | None=None,
) -> Tuple[float, float]:
    """Determine how far along a focal coordinate system axis a DataArray's value stays above a certain cutoff.

    See `get_focus_matrix` for the meaning of "focal coordinate system."

    Args:
        da: DataArray whose values will be considered (presumably defined on transducer coordinates).
        focus: A 3D point describing the focus location in the coordinates of `da`
        dim: The name of the dimension of `da` whose corresponding focal coordinate system axis should be sampled along.
            For example, the "axial" dimension of a transducer corresponds to the focal axis (z-axis) in the focal coordinate
            system, that is, the ray from the transducer's "effective origin" (see `Transducer.get_effective_origin`)
            to the focus center.
        cutoff: The threshold against which `da` values are compared.
        origin: A 3D point describing the "effective origin" in the coordinates of `da`
            (see `Transducer.get_effective_origin` for the meaning of this).
        min_offset: How far along the negative focal `dim` direction to sample. By default samples as far as the coordinate
            grid allows.
        max_offset: How far along the positive focal `dim` direction to sample. By default samples as far as the coordinate
            grid allows.

    Returns:
        negoff: Distance from the `focus` along the negative focal `dim` direction until `da` first goes below `cutoff`.
        posoff: Distance from the `focus` along the positive focal `dim` direction until `da` first goes below `cutoff`.
    """
    interp_da = interp_transformed_axis(da, focus=focus, dim=dim, origin=origin, min_offset=min_offset, max_offset=max_offset)
    offset = interp_da.coords[f'offset_d{dim}']
    da_negoff = interp_da.where(offset <= 0, drop=True)
    da_posoff = interp_da.where(offset >= 0, drop=True)
    da_negoff = da_negoff.where(da_negoff < float(cutoff), drop=True)
    if da_negoff.size > 0:
        negoff = float(da_negoff.coords[f'offset_d{dim}'][-1])
    else:
        negoff = np.nan
    da_posoff = da_posoff.where(da_posoff < float(cutoff), drop=True)
    if da_posoff.size > 0:
        posoff = float(da_posoff.coords[f'offset_d{dim}'][0])
    else:
        posoff = np.nan
    return negoff, posoff

def get_beamwidth(
    da: xa.DataArray,
    focus,
    dim,
    cutoff:float | None=None,
    origin=DEFAULT_ORIGIN,
    min_offset:float | None=None,
    max_offset:float | None=None
) -> float:
    """Determine the FWHM (or differently thresholded width) of a DataArray along a focal coordinate system axis.

    See `get_focus_matrix` for the meaning of "focal coordinate system."

    Args:
        da: DataArray whose values will be considered (presumably defined on transducer coordinates).
        focus: A 3D point describing the focus location in the coordinates of `da`
        dim: The name of the dimension of `da` whose corresponding focal coordinate system axis should be sampled along.
            For example, the "axial" dimension of a transducer corresponds to the focal axis (z-axis) in the focal coordinate
            system, that is, the ray from the transducer's "effective origin" (see `Transducer.get_effective_origin`)
            to the focus center.
        cutoff: The threshold against which `da` values are compared. If not provided then the half-max is used, making
            this an "FWHM" function.
        origin: A 3D point describing the "effective origin" in the coordinates of `da`
            (see `Transducer.get_effective_origin` for the meaning of this).
        min_offset: How far along the negative focal `dim` direction to sample. By default samples as far as the coordinate
            grid allows.
        max_offset: How far along the positive focal `dim` direction to sample. By default samples as far as the coordinate
            grid allows.

    Returns: The "beam width" along the focal `dim` direction, by measuring from the from closest-to-focus point along the
        *negative* focal `dim` axis for which `da` goes below `cutoff` up to the closest-to-focus point along the
        *positive* focal `dim` axis for which `da` goes below `cutoff`.
    """
    if cutoff is None:
        cutoff = float(da.max())/2
    negoff, posoff = get_beam_bounds(da, focus=focus, dim=dim, cutoff=float(cutoff), origin=origin, min_offset=min_offset, max_offset=max_offset)
    bw = posoff - negoff
    return bw



from __future__ import annotations

import pandas as pd
import xarray as xa

from openlifu.seg.material import MATERIALS, Material
from openlifu.seg.seg_method import SegmentationMethod


class UniformSegmentation(SegmentationMethod):
    def _segment(self, volume: xa.DataArray):
        return self._ref_segment(volume.coords)

    def to_table(self) -> pd.DataFrame:
        """
        Get a table of the segmentation method parameters

        :returns: Pandas DataFrame of the segmentation method parameters
        """
        records = [{"Name": "Type", "Value": "Uniform", "Unit": ""},
                   {"Name": "Reference Material", "Value": self.ref_material, "Unit": ""}]
        return pd.DataFrame.from_records(records)

class UniformTissue(UniformSegmentation):
    """ Assigns the tissue material to all voxels in the volume. """
    def __init__(self, materials: dict[str, Material] | None = None):
        if materials is None:
            materials = MATERIALS.copy()
        super().__init__(materials=materials, ref_material="tissue")

    def to_table(self) -> pd.DataFrame:
        """
        Get a table of the segmentation method parameters

        :returns: Pandas DataFrame of the segmentation method parameters
        """
        records = [{"Name": "Type", "Value": "Uniform Tissue", "Unit": ""}]
        return pd.DataFrame.from_records(records)

    def to_dict(self):
        d = super().to_dict()
        d.pop("ref_material")
        return d


class UniformWater(UniformSegmentation):
    """ Assigns the water material to all voxels in the volume. """
    def __init__(self, materials: dict[str, Material] | None = None):
        if materials is None:
            materials = MATERIALS.copy()
        super().__init__(materials=materials, ref_material="water")

    def to_table(self) -> pd.DataFrame:
        """
        Get a table of the segmentation method parameters

        :returns: Pandas DataFrame of the segmentation method parameters
        """
        records = [{"Name": "Type", "Value": "Uniform Water", "Unit": ""}]
        return pd.DataFrame.from_records(records)

    def to_dict(self):
        d = super().to_dict()
        d.pop("ref_material")
        return d



from __future__ import annotations

import sys
import time

from openlifu.io.LIFUInterface import LIFUInterface

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/demo.py

def get_user_input():
    while True:
        print("\nEnter parameters (or 'x' to exit):")
        try:
            freq = input("Trigger Frequency (Hz): ")
            if freq.lower() == 'x':
                return None

            pulse_width = input("Pulse Width (μs): ")
            if pulse_width.lower() == 'x':
                return None

            return {
                "freq": float(freq),
                "pulse_width": float(pulse_width)
            }
        except ValueError:
            print("Invalid input. Please enter numbers or 'x' to exit.")

def main():
    print("Starting LIFU Test Script...")
    interface = LIFUInterface()
    tx_connected, hv_connected = interface.is_device_connected()

    if hv_connected:
        console_version = interface.hvcontroller.get_version()
        print(f"Version: {console_version}")
        print("HV Controller connected.")
        interface.hvcontroller.set_voltage(24.5)
        interface.hvcontroller.turn_hv_on()
        input("Press Enter to continue...")

        interface.hvcontroller.turn_hv_off()



    if not tx_connected and not hv_connected:
        print("✅ LIFU Console not connected.")
        sys.exit(1)

    if not tx_connected:
        print("TX device not connected. Attempting to turn on 12V...")
        interface.hvcontroller.turn_12v_on()
        time.sleep(2)

        interface.stop_monitoring()
        del interface
        time.sleep(3)

        print("Reinitializing LIFU interface after powering 12V...")
        interface = LIFUInterface()
        tx_connected, hv_connected = interface.is_device_connected()

        if tx_connected and hv_connected:
            print("✅ LIFU Device fully connected.")
        else:
            print("❌ LIFU Device NOT fully connected.")
            print(f"  TX Connected: {tx_connected}")
            print(f"  HV Connected: {hv_connected}")
            sys.exit(1)

        print("HV Controller connected.")
        interface.hvcontroller.set_voltage(24.5)
        interface.hvcontroller.turn_hv_on()

    device_count = interface.txdevice.get_tx_module_count()
    print(f"Device Count: {device_count}")

    for i in range(1, device_count + 1):

        print("Ping the device")
        if not interface.txdevice.ping(module=i):
            print("❌ Failed comms with txdevice.")
            sys.exit(1)

        version = interface.txdevice.get_version(module=i)
        print(f"Version: {version}")


    if hv_connected:
        time.sleep(5)
        print("disabling HV for safety...")
        interface.hvcontroller.turn_hv_off()

if __name__ == "__main__":
    main()



from __future__ import annotations

import pytest

from openlifu.plan.param_constraint import ParameterConstraint
from openlifu.plan.solution_analysis import SolutionAnalysis, SolutionAnalysisOptions

# ---- Tests for SolutionAnalysis ----

@pytest.fixture()
def example_solution_analysis() -> SolutionAnalysis:
    return SolutionAnalysis(
        mainlobe_pnp_MPa=[1.1, 1.2],
        mainlobe_isppa_Wcm2=[10.0, 12.0],
        mainlobe_ispta_mWcm2=[500.0, 520.0],
        beamwidth_lat_3dB_mm=[1.5, 1.6],
        beamwidth_ele_3dB_mm=[2.0, 2.1],
        beamwidth_ax_3dB_mm=[3.0, 3.1],
        beamwidth_lat_6dB_mm=[1.8, 1.9],
        beamwidth_ele_6dB_mm=[2.5, 2.6],
        beamwidth_ax_6dB_mm=[3.5, 3.6],
        sidelobe_pnp_MPa=[0.5, 0.6],
        sidelobe_isppa_Wcm2=[5.0, 5.5],
        sidelobe_to_mainlobe_pressure_ratio=[0.5/1.1, 0.6/1.2], # approx 0.45, 0.5
        sidelobe_to_mainlobe_intensity_ratio=[5.0/10.0, 5.5/12.0], # 0.5, approx 0.458
        global_pnp_MPa=[1.3],
        global_isppa_Wcm2=[13.0],
        p0_MPa=[1.0, 1.1],
        TIC=0.7,
        power_W=25.0,
        MI=1.2,
        global_ispta_mWcm2=540.0,
        param_constraints={
            "global_pnp_MPa": ParameterConstraint(
                operator="<=",
                warning_value=1.4,
                error_value=1.6
            )
        }
    )

def test_to_dict_from_dict_solution_analysis(example_solution_analysis: SolutionAnalysis):
    sa_dict = example_solution_analysis.to_dict()
    new_solution = SolutionAnalysis.from_dict(sa_dict)
    assert new_solution == example_solution_analysis

@pytest.mark.parametrize("compact", [True, False])
def test_serialize_deserialize_solution_analysis(example_solution_analysis: SolutionAnalysis, compact: bool):
    json_str = example_solution_analysis.to_json(compact)
    deserialized = SolutionAnalysis.from_json(json_str)
    assert deserialized == example_solution_analysis


# ---- Tests for SolutionAnalysisOptions ----


@pytest.fixture()
def example_solution_analysis_options() -> SolutionAnalysisOptions:
    return SolutionAnalysisOptions(
        standoff_sound_speed=1480.0,
        standoff_density=990.0,
        ref_sound_speed=1540.0,
        ref_density=1020.0,
        mainlobe_aspect_ratio=(1.0, 1.0, 4.0),
        mainlobe_radius=2.0e-3,
        beamwidth_radius=4.0e-3,
        sidelobe_radius=2.5e-3,
        sidelobe_zmin=0.5e-3,
        distance_units="mm",
        param_constraints={
            "mainlobe_radius": ParameterConstraint(
                operator=">=",
                warning_value=1.5e-3,
                error_value=1.0e-3
            )
        }
    )

def test_to_dict_from_dict_solution_analysis_options(example_solution_analysis_options: SolutionAnalysisOptions):
    options_dict = example_solution_analysis_options.to_dict()
    new_options = SolutionAnalysisOptions.from_dict(options_dict)
    assert new_options == example_solution_analysis_options



from __future__ import annotations

import datetime
import importlib
import json
import logging
import re
import shutil
import tempfile
import threading
import time
from dataclasses import dataclass
from pathlib import Path
from subprocess import PIPE, CalledProcessError, CompletedProcess, Popen
from typing import Annotated, Any, Callable, Dict, List, Tuple

import cv2
import numpy as np
import onnxruntime as ort
import OpenEXR
import trimesh
import vtk
from PIL import Image
from vtk.util.numpy_support import numpy_to_vtk

from openlifu.util.annotations import OpenLIFUFieldData
from openlifu.util.assets import download_and_install_modnet, get_modnet_path

logger_meshrecon = logging.getLogger("MeshRecon")
logger_meshroom = logging.getLogger("Meshroom")

@dataclass
class Photoscan:
    id: Annotated[str, OpenLIFUFieldData("Photoscan ID", "ID of this photoscan")] = "photoscan"
    """ID of this photoscan"""

    name: Annotated[str, OpenLIFUFieldData("Photoscan name", "Photoscan name")] = "Photoscan"
    """Photoscan name"""

    model_filename: Annotated[str | None, OpenLIFUFieldData("Model filename", "Relative path to model")] = None
    """Relative path to model"""

    texture_filename: Annotated[str | None, OpenLIFUFieldData("Texture filename", "Relative path to texture image")] = None
    """Relative path to texture image"""

    mtl_filename: Annotated[str | None, OpenLIFUFieldData("Material filename", "Relative path to materials file")] = None
    """Relative path to materials file"""

    photoscan_approved: Annotated[bool, OpenLIFUFieldData("Approved?", "Approval state of the photoscan. 'True' means the user has provided some kind of confirmation that the photoscan is good enough to be used.")] = False
    """Approval state of the photoscan. 'True' means the user has provided some kind of
    confirmation that the photoscan is good enough to be used."""

    @staticmethod
    def from_json(json_string: str) -> Photoscan:
        """Load a Photoscan from a json string"""
        return Photoscan.from_dict(json.loads(json_string))

    def to_json(self, compact: bool) -> str:
        """Serialize a Photoscan to a json string. This is different to the format written to file
        and does not contain the loaded models and texture.
        Args:
            compact:if enabled then the string is compact (not pretty). Disable for pretty.
        Returns: A json string representing the complete Photoscan object
        """
        from openlifu.util.json import PYFUSEncoder

        if compact:
            return json.dumps(self.to_dict(), separators=(',', ':'), cls=PYFUSEncoder)
        else:
            return json.dumps(self.to_dict(), indent=4, cls=PYFUSEncoder)

    @staticmethod
    def from_dict(d:Dict) -> Photoscan:
        """
        Create a Photoscan from a dictionary
        param d: Dictionary of photoscan parameters.
        returns: Photoscan object
        """
        return Photoscan(**d)

    def to_dict(self) -> Dict:
        """
        Convert the photoscan to a dictionary
        returns: Dictionary of photoscan parameters
        """
        d = self.__dict__.copy()
        return d

    @staticmethod
    def from_file(filename) -> Photoscan:
        """
        Load a Photoscan from a metadata file
        :param filename: Name of the file
        """
        with open(filename) as f:
            return Photoscan.from_dict(json.load(f))

    def to_file(self, filename):
        """
        Save the photoscan to a file.
        :param filename: Name of the file
        """
        Path(filename).parent.parent.mkdir(exist_ok=True) #photoscan directory
        Path(filename).parent.mkdir(exist_ok=True)
        with open(filename, 'w') as file:
            file.write(self.to_json(compact = False))

def load_data_from_filepaths(model_abspath: str, texture_abspath: str | None) -> Tuple[vtk.vtkPolyData, vtk.vtkImageData | None]:
    """
    This function returns the data directly from the model and texture filepaths without requiring a photoscan object.
    param model_abspath: absolute filepath to model data
          texture abspath: absolute filepath to texture data
    Returns: Photoscan data as (model_vtkpolydata, texture_vtkimagedata)
    """
    model_polydata  = load_model(model_abspath)
    texture_imagedata = None
    if texture_abspath:
        texture_imagedata = load_texture(texture_abspath)

    return (model_polydata, texture_imagedata)

def load_data_from_photoscan(photoscan: Photoscan, parent_dir) -> Tuple[vtk.vtkPolyData, vtk.vtkImageData | None]:
    """
    param parent_dir: parent directory containing model and texture data filepaths
    Returns: Photoscan data as (model_vtkpolydata, texture_vtkimagedata)
    """
    model_polydata  = load_model(Path(parent_dir)/photoscan.model_filename)
    texture_imagedata = None
    if photoscan.texture_filename:
        texture_imagedata = load_texture(Path(parent_dir)/photoscan.texture_filename)

    return (model_polydata, texture_imagedata)

def load_model(file_name) -> vtk.vtkPolyData:
    """
    This function assumes that the model is saved to file in LPS coordinates.
    The model that is returned is in RAS coordinates
    """
    if not Path(file_name).exists():
        raise FileNotFoundError(f'Model filepath does not exist: {file_name}')
    mesh = read_as_vtkpolydata(file_name)
    mesh_ras = convert_between_ras_and_lps(mesh)

    return mesh_ras

def load_texture(file_name) -> vtk.vtkImageData:
    if not Path(file_name).exists():
        raise FileNotFoundError(f'Texture data filepath does not exist: {file_name}')
    texture = read_as_vtkimagedata(file_name)
    return texture

def read_as_vtkpolydata(file_name):

    suffix_to_reader_dict = {
        '.ply' : vtk.vtkPLYReader,
        '.vtp' : vtk.vtkXMLPolyDataReader,
        '.obj' : vtk.vtkOBJReader,
        '.stl' : vtk.vtkSTLReader,
        '.vtk' : vtk.vtkPolyDataReader,
        '.g' : vtk.vtkBYUReader
    }
    path = Path(file_name)
    ext = path.suffix.lower()
    if path.suffix not in suffix_to_reader_dict:
        raise ValueError(f"File format {ext} not supported by reader")
    reader = suffix_to_reader_dict[ext]()
    if ext == '.g':
        reader.SetGeometryName(file_name)
    else:
        reader.SetFileName(file_name)
    reader.Update()
    poly_data = reader.GetOutput()

    return poly_data

def convert_numpy_to_vtkimage(image_numpy: np.ndarray) -> vtk.vtkImageData:
    """
    Converts a numpy array with dimensions [HxWx3] representing an RGB image into vtkImageData
    """
    vtkimage_data = vtk.vtkImageData()
    vtkimage_data.SetDimensions(image_numpy.shape[1], image_numpy.shape[0], 1)
    vtkimage_data.SetNumberOfScalarComponents(image_numpy.shape[2], vtkimage_data.GetInformation())
    pd = vtkimage_data.GetPointData()
    new_rgb_data = image_numpy.reshape((-1, image_numpy.shape[2]))
    vtk_array = numpy_to_vtk(new_rgb_data, deep=True, array_type=vtk.VTK_UNSIGNED_SHORT)
    pd.SetScalars(vtk_array)
    return vtkimage_data

def read_as_vtkimagedata(file_name) -> vtk.vtkImageData:

    suffix_to_reader_dict = {
        '.jpg' : vtk.vtkJPEGReader,
        '.png' : vtk.vtkPNGReader,
        '.tiff' : vtk.vtkTIFFReader,
    }
    path = Path(file_name)
    ext = path.suffix.lower()
    if ext == '.exr':
        with OpenEXR.File(str(file_name), separate_channels = True) as exr_file:
            R = exr_file.channels()['R'].pixels
            G = exr_file.channels()['G'].pixels
            B = exr_file.channels()['B'].pixels
            # Combine channels into a single RGB image (H x W x 3)
            rgb_data = np.stack([R, G, B], axis=-1, dtype = np.float32)
            # EXR stores pixel data using half 16-bit floating point values. Convert data to an 8-bit image compatibility with VTK
            rgb_data = np.clip(rgb_data*(2**16-1), 0, 65535)
            image_data = convert_numpy_to_vtkimage(rgb_data)
    elif ext in suffix_to_reader_dict:
        reader = suffix_to_reader_dict[ext]()
        reader.SetFileName(file_name)
        reader.Update()
        image_data = reader.GetOutput()
    else:
        raise ValueError(f"File format {path.suffix} not supported by reader")

    return image_data

def convert_between_ras_and_lps(mesh : vtk.vtkPointSet) -> vtk.vtkPointSet:
    """Converts a mesh (polydata, unstructured grid or even just a point cloud) between the LPS (left-posterior-superior) coordinate system and
      RAS (right-anterior-superior) coordinate system."""

    transform_ras_to_lps = vtk.vtkTransform()
    transform_ras_to_lps.Scale(-1,-1,1)

    transformFilter = vtk.vtkTransformPolyDataFilter()
    transformFilter.SetInputData(mesh)
    transformFilter.SetTransform(transform_ras_to_lps)
    transformFilter.Update()

    return transformFilter.GetOutput()

def get_meshroom_pipeline_names() -> list[str]:
    """Get a list of names of valid meshroom pipelines that can be used in run_reconstruction"""
    pipeline_dir = importlib.resources.files("openlifu.nav.meshroom_pipelines")
    return [f.stem for f in pipeline_dir.iterdir() if f.suffix == ".mg"]

def subprocess_stream_output(
    args: str | List[str],
    stdout_handler: Callable[[str], None],
    stderr_handler: Callable[[str], None],
    check: bool = True,
    text: bool = True,
    **kwargs: Any,
) -> CompletedProcess:
    """
    Run a subprocess and stream its stdout and stderr output to separate handlers/loggers. A drop in replacement
    for subprocess.run that handles logging.

    Args:
        args (Union[str, List[str]]): Command and arguments to execute. Can be a string or list of strings.
        stdout_handler (Callable[[str], None]): Function to handle each line of standard output.
        stderr_handler (Callable[[str], None]): Function to handle each line of standard error.
        check (bool, optional): If True, raise CalledProcessError if the subprocess exits with a non-zero code.
        text (bool, optional): If True, communicate with the process using text mode. Defaults to True.
        **kwargs (Any): Additional keyword arguments passed to subprocess.Popen.

    Returns:
        CompletedProcess: An object containing the arguments used and the return code.

    Raises:
        CalledProcessError: If `check` is True and the subprocess exits with a non-zero status.
    """
    with Popen(args, stdout=PIPE, stderr=PIPE, text=text, **kwargs) as process:

        def log_stream(stream, handler):
            for line in stream:
                handler(line.rstrip())

        # Create threads for each stream
        stdout_thread = threading.Thread(target=log_stream, args=(process.stdout, stdout_handler))
        stderr_thread = threading.Thread(target=log_stream, args=(process.stderr, stderr_handler))
        stdout_thread.start()
        stderr_thread.start()

        stdout_thread.join()
        stderr_thread.join()

        retcode = process.wait()

    if check and retcode:
        raise CalledProcessError(retcode, process.args)
    return CompletedProcess(process.args, retcode)

#Just hard code node order
_nodes = [  "FeatureExtraction_1",
            "FeatureMatching_1",
            "StructureFromMotion_1",
            "PrepareDenseScene_1",
            "DepthMap_1",
            "DepthMapFilter_1",
            "Meshing_1",
            "MeshFiltering_1",
            "Texturing_1",
            "Publish_1" ]

def run_reconstruction(
    images: list[Path],
    pipeline_name: str = "default_pipeline",
    input_resize_width: int = 3024,
    use_masks: bool = True,
    matching_mode: str = 'exhaustive',
    window_radius: int | None = None,
    num_neighbors: int | None = None,
    locations: List[Tuple[float, float, float]] | None = None,
    return_durations: bool = False,
    progress_callback : Callable[[int,str],None] | None = None,
    download_masking_model: bool = True,
) -> Tuple[Photoscan, Path] | Tuple[Photoscan, Path, Dict[str, float]]:
    """Run Meshroom with the given images and pipeline.
    Args:
        images (list[Path]): List of image file paths.
        pipeline_name (str): Name of the Meshroom pipeline in meshroom_pipelines folder.
            See also `get_meshroom_pipeline_names`.
        input_resize_width (int): Width to which input images will be resized, in pixels.
        use_masks (bool): Whether to include a background removal step to filter the dense reconstruction.
        matching_mode (str): Strategy for generating image pairs. One of:
            - 'exhaustive': Match every image with every other image.
            - 'sequential': Match each image with the previous and next `window_radius` images (no wrap-around).
            - 'sequential_loop': Like 'sequential' but wraps around at the end of the sequence.
            - 'spatial': Match each image with its `num_neighbors` nearest neighbors based on 3D location.
        window_radius (int | None): Required for 'sequential' and 'sequential_loop' matching_mode. Number of
            images forward and backward in the sequence to try and match with.
        num_neighbors (int | None): Required for 'spatial' matching_mode. Number of nearest neighbors to match
            based on 3D distance in `locations`.
        locations (List[Tuple[float, float, float]]] | None): Required for 'spatial'. Must be the same length
            as `images`. Provides 3D coordinates for spatial matching.
        return_durations (bool): If True, also return a dictionary mapping node names to durations in seconds.
        progress_callback: An optional function that will be called to report progress. The function should accept two arguments:
            an integer progress value from 0 to 100 followed by a string message describing the step currently being worked on.
        download_masking_model: Whether to auto-download the masking model weights if they are not present;
            only relevant if use_masks is enabled.

    Returns:
        Union[Tuple[Photoscan, Path], Tuple[Photoscan, Path, Dict[str, float]]]:
            - If return_durations is False: returns the Photoscan and the data directory path.
            - If return_durations is True: also returns a dictionary of node execution times.
    """
    image_indices = list(range(len(images)))
    valid_modes = {'exhaustive', 'sequential', 'sequential_loop', 'spatial'}
    if matching_mode == 'exhaustive':
        pairs = _make_pairs_sequential(image_indices, len(image_indices))

    elif matching_mode == 'sequential':
        if window_radius is None:
            raise ValueError(f"A window radius is required for matching mode: '{matching_mode}'.")
        pairs = _make_pairs_sequential(image_indices, window_radius)

    elif matching_mode == 'sequential_loop':
        if window_radius is None:
            raise ValueError(f"A window radius is required for matching mode: '{matching_mode}'.")
        pairs = _make_pairs_sequential_loop(image_indices, window_radius)

    elif matching_mode == 'spatial':
        if locations is None:
            raise ValueError("Spatial matching requires `locations`, but none were provided.")
        if len(locations) != len(images):
            raise ValueError("`locations` must be the same length as `images`.")
        if num_neighbors is None:
            raise ValueError("Spatial matching requires `num_neighbors`, but it was not provided.")
        pairs = _make_pairs_spatial(image_indices, num_neighbors, locations)

    else:
        raise ValueError(f"Invalid matching mode: '{matching_mode}'. Must be one of {valid_modes}.")

    if progress_callback is None:
        def progress_callback(progress_percent : int, step_description : str): # noqa: ARG001
            pass # Define it to be a no-op if no progress_callback was provided.

    progress_callback(0, "Starting reconstruction")

    durations = {}

    pipeline_dir = importlib.resources.files("openlifu.nav.meshroom_pipelines")
    valid_configs = get_meshroom_pipeline_names()

    if pipeline_name not in valid_configs:
        raise ValueError(
            f"Invalid pipeline name '{pipeline_name}'. "
            f"Valid options are: {', '.join(valid_configs)}"
        )

    pipeline = pipeline_dir / f"{pipeline_name}.mg"

    config_nodes, downscale_factor = get_meshroom_config_info(pipeline)

    if shutil.which("meshroom_batch") is None:
        raise FileNotFoundError("Error: 'meshroom_batch' is not found in system PATH. Ensure it is installed and accessible.")

    temp_dir = Path(tempfile.mkdtemp())

    images_dir = temp_dir / "images"
    images_dir.mkdir(parents=True, exist_ok=True)

    #resize the images and store in tmp dir
    progress_callback(0, "Resizing images")
    start_time = time.perf_counter()
    new_paths = []
    for image in images:
        img = Image.open(image)
        orientation = img.getexif().get(274, 1)
        if orientation in range(4,8):
            #camera is rotated. Get the perceived width.
            old_width = img.height
        else:
            old_width = img.width
        scale = input_resize_width/old_width

        resize_width = int(scale*img.width)
        resize_height = int(scale*img.height)
        min_allowed = downscale_factor * 128
        min_side = min(resize_height, resize_width)

        if downscale_factor > 0 and min_side < min_allowed:
            raise ValueError(
                f"Minimum resized side length is {min_side}, but the minimum allowed for pipeline '{pipeline_name}' is {min_allowed}. "
                "Please increase `input_resize_width` or use a pipeline with less downscaling."
            )

        img = img.resize((resize_width, resize_height), Image.Resampling.LANCZOS)
        img.save( images_dir / image.name, exif=img.getexif())
        new_paths.append(images_dir / image.name)
    durations["ImageResizing"] = time.perf_counter() - start_time

    output_dir = temp_dir / "output"
    cache_dir = temp_dir / "cache"
    pair_file_path = cache_dir / "imageMatches.txt"

    command = [
        "meshroom_batch",
        "--pipeline", pipeline.as_posix(),
        "--output", output_dir.as_posix(),
        "--input", images_dir.as_posix(),
        "--cache", cache_dir.as_posix(),
        "--paramOverrides", f"FeatureMatching_1.imagePairsList={pair_file_path.as_posix()}",
    ]

    if use_masks:
        progress_callback(3, "Generating image masks")
        start_time = time.perf_counter()
        masks_dir = temp_dir / "masks"
        masks_dir.mkdir(parents=True, exist_ok=True)
        make_masks(new_paths, masks_dir, download_model=download_masking_model)
        command.append( f"PrepareDenseScene_1.masksFolders=['{masks_dir.as_posix()}']" )
        durations["MaskCreation"] = time.perf_counter() - start_time

    #run CameraInit node to set view ids
    progress_callback(8, "Initializing camera IDs")
    command_camera_init = [*command.copy(), "--toNode", "CameraInit"]
    subprocess_stream_output(command_camera_init, logger_meshroom.info, logger_meshroom.warning)

    camera_init_path = next(cache_dir.glob("CameraInit/*/cameraInit.sfm"))
    write_pair_file(new_paths, pairs, camera_init_path, pair_file_path)

    number_of_nodes = len([node for node in _nodes if node in config_nodes])
    pipeline_progress_start = 10.0
    pipeline_progress_end = 90.0
    pipeline_progress_step = (pipeline_progress_end - pipeline_progress_start) / number_of_nodes

    #run each node
    pipeline_progress = pipeline_progress_start
    for node in _nodes:
        if node not in config_nodes:
            continue
        command_i = [*command.copy(), "--toNode", node]

        progress_callback(int(pipeline_progress),f"Meshroom: {node.strip('_1')}")
        start_time = time.perf_counter()
        subprocess_stream_output(command_i, logger_meshroom.info, logger_meshroom.warning)
        durations[node] = time.perf_counter() - start_time
        pipeline_progress += pipeline_progress_step

    progress_callback(pipeline_progress_end, "Merging texture maps")

    #merge texturemaps
    output_dir_merged = temp_dir / "output_dir_merged"
    output_dir_merged.mkdir(parents=True, exist_ok=True)

    merge_textures(output_dir / "texturedMesh.obj", output_dir_merged)

    photoscan_dict = {
        "model_filename":  "texturedMesh.obj",
        "texture_filename": "material_0.png",
        "mtl_filename": "material.mtl",
    }

    photoscan = Photoscan.from_dict(photoscan_dict)
    progress_callback(100, "Complete")
    if return_durations:
        return photoscan, output_dir_merged, durations
    return photoscan, output_dir_merged

def udim_to_tile(udim_str: str) -> Tuple[int, int]:
    """Convert UDIM string to tile coordinates"""
    x = int(udim_str[-2:]) - 1
    tile_u = x % 10
    tile_v = x // 10
    return tile_u, tile_v

def merge_textures(input_obj_path: Path, output_path: Path) -> None:
    """Merge the UDIM textures output by meshroom into a single large texture

    Args:
        input_obj_path (Path): Path to .obj mesh to merge.
        output_path (Path): Path to save merged .obj mesh.
    """
    scene = trimesh.load(input_obj_path, process=True)

    if isinstance(scene, trimesh.Scene):
        mesh_dict = scene.geometry
    else:
        mesh_dict = {'material_1001': scene}

    for name, mesh in mesh_dict.items():
        mesh.tile = udim_to_tile(name[-4:])

    num_u_tiles = max(mesh.tile[0] for _, mesh in mesh_dict.items()) + 1
    num_v_tiles = max(mesh.tile[1] for _, mesh in mesh_dict.items()) + 1

    first_tile = np.array(mesh_dict['material_1001'].visual.material.image)
    tile_height, tile_width, tile_ch = first_tile.shape
    tex_height, tex_width = num_v_tiles*tile_height, num_u_tiles*tile_height
    new_texture = np.zeros((tex_height, tex_width, tile_ch), dtype=first_tile.dtype)

    new_verts = []
    new_faces = []
    new_uvs = []
    num_verts = 0
    for _, mesh in mesh_dict.items():
        tile_u, tile_v = mesh.tile
        texture_tile = np.array(mesh.visual.material.image)
        new_texture[tex_height - tile_height*(tile_v+1): tex_height-tile_height*tile_v, tile_width*tile_u:tile_width*(tile_u+1)] = texture_tile
        new_verts.append(mesh.vertices)
        new_faces.append(mesh.faces+num_verts)
        new_uvs.append(mesh.visual.uv)
        num_verts += mesh.vertices.shape[0]

    new_texture = Image.fromarray(new_texture)
    new_verts = np.concatenate(new_verts, axis=0)
    new_faces = np.concatenate(new_faces, axis=0)
    new_uvs = np.concatenate(new_uvs, axis=0)

    new_uvs = new_uvs/np.array([num_u_tiles, num_v_tiles]).reshape(-1,2)

    mesh = trimesh.Trimesh() #Note putting vertices in this constructor will merge nearby ones
    mesh.vertices = new_verts
    mesh.faces = new_faces
    mesh.visual = trimesh.visual.TextureVisuals(uv=new_uvs, image=new_texture)

    mesh.export(output_path / "texturedMesh.obj")


def apply_exif_orientation_numpy(image: np.ndarray, orientation: int, inverse: bool = False) -> np.ndarray:
    """
    Transforms an image array to undo or redo EXIF orientation.

    Args:
        image (np.ndarray): The image array (H x W x C or H x W).
        orientation (int): EXIF orientation tag value (1-8).
        inverse (bool): If True, applies the inverse transformation.

    Returns:
        np.ndarray: The transformed image.
    """
    if orientation <= 1 or orientation > 8:
        return image  # No transformation needed

    # Define forward transformations
    def transform(img):
        if orientation == 2:
            return np.fliplr(img)
        elif orientation == 3:
            return np.rot90(img, 2)
        elif orientation == 4:
            return np.flipud(img)
        elif orientation == 5:
            return np.rot90(np.fliplr(img), 1)
        elif orientation == 6:
            return np.rot90(img, -1)
        elif orientation == 7:
            return np.rot90(np.fliplr(img), -1)
        elif orientation == 8:
            return np.rot90(img, 1)
        else:
            raise ValueError("Unknown exif orientation")

    # Define inverse transformations
    def inverse_transform(img):
        if orientation == 2:
            return np.fliplr(img)
        elif orientation == 3:
            return np.rot90(img, 2)
        elif orientation == 4:
            return np.flipud(img)
        elif orientation == 5:
            return np.fliplr(np.rot90(img, -1))
        elif orientation == 6:
            return np.rot90(img, 1)
        elif orientation == 7:
            return np.fliplr(np.rot90(img, 1))
        elif orientation == 8:
            return np.rot90(img, -1)
        else:
            raise ValueError("Unknown exif orientation")

    return inverse_transform(image) if inverse else transform(image)

def preprocess_image_modnet(image: np.ndarray, ref_size: int = 512) -> np.ndarray:
    """
    Preprocess an input image for MODNet inference.

    This function performs the same preprocessing steps as the official MODNet code:
    - Normalizes image values to the range [-1, 1]
    - Resizes the image based on a reference size (512), maintaining aspect ratio
    - Ensures the resized dimensions are divisible by 32
    - Converts the image to CHW format and adds a batch dimension

    Args:
        image (np.ndarray): Input image in HWC format with values in [0, 255].

    Returns:
        np.ndarray: Preprocessed image in NCHW format with float32 values in [-1, 1].
    """

    # Normalize image to [-1, 1]
    image = image.astype(np.float32) / 255.0
    image = 2 * image - 1

    im_h, im_w, _ = image.shape

    # Resize if both dimensions are smaller or larger than the reference size
    if max(im_h, im_w) < ref_size or min(im_h, im_w) > ref_size:
        if im_w >= im_h:
            im_rh = ref_size
            im_rw = int(im_w * ref_size / im_h)
        else:
            im_rw = ref_size
            im_rh = int(im_h * ref_size / im_w)
    else:
        im_rh, im_rw = im_h, im_w

    # Ensure dimensions are divisible by 32
    im_rw -= im_rw % 32
    im_rh -= im_rh % 32

    # Resize, convert to CHW, and add batch dimension
    image = cv2.resize(image, (im_rw, im_rh), interpolation=cv2.INTER_AREA)
    image = np.transpose(image, (2, 0, 1))  # HWC to CHW
    image = np.expand_dims(image, axis=0)  # Add batch dimension

    return image


def make_masks(image_paths: list[Path], output_dir: Path, threshold: float = 0.01, download_model=True) -> None:
    """
    Runs MODNet on a list of image paths and saves the output masks.

    Each output mask is saved in `output_dir` using the original filename with a `.png` extension.
    The `threshold` parameter is used to convert the MODNet soft segmentation output into a binary (hard) mask.
    EXIF orientation data is preserved to ensure correct image alignment when loading into Meshroom.

    Args:
        image_paths (List[str]): List of input image file paths.
        output_dir (str): Directory where the output masks will be saved.
        threshold (float): Threshold to binarize the soft segmentation output.
        download_model (bool): Whether to auto-download the model weights if they are not present.
    """

    # Load the ONNX model

    ckpt_path = get_modnet_path()
    if not ckpt_path.exists():
        if download_model:
            logger_meshrecon.info(f"Downloading MODNet checkpoint to {ckpt_path}")
            download_and_install_modnet()
        else:
            raise FileNotFoundError(f"MODNet checkpoint not found at {ckpt_path}. Install it using an appropirate utility in openlifu.util.assets.")
    else:
        logger_meshrecon.info(f"Found existing MODNet checkpoint at {ckpt_path}")

    session = ort.InferenceSession(ckpt_path, providers=["CPUExecutionProvider"])  # or CUDAExecutionProvider

    for image_path in image_paths:
        image = Image.open(image_path)
        exif = image.getexif()
        orientation = exif.get(274,1)
        image = np.array(image)
        image = apply_exif_orientation_numpy(image, orientation=orientation)
        im_h, im_w, _ = image.shape
        image = preprocess_image_modnet(image)

        # Run the model
        mask = session.run(["output"], {"input": image})[0].squeeze()
        mask = cv2.resize(mask, (im_w, im_h), interpolation=cv2.INTER_AREA)
        mask = mask > threshold
        mask = (mask.astype(np.float32)*255).astype(np.uint8)
        mask = np.repeat(mask[:, :, np.newaxis], 3, axis=2)

        mask = apply_exif_orientation_numpy(mask, orientation=orientation, inverse=True)

        mask = Image.fromarray(mask)
        output_path = output_dir / (image_path.stem + ".png")
        mask.save(output_path, exif=exif)

def _read_path_to_view_ids(input_path: Path) -> Dict[Path, str]:
    """Read the mapping from image paths to viewIds produced by Meshroom CameraInit node"""
    with open(input_path) as f:
        data = json.load(f)

    basename_to_viewid = {
        Path(view['path']): view['viewId']
        for view in data.get('views', [])
    }
    return basename_to_viewid


def _make_pairs_sequential_loop(view_ids: List[Any], window_radius: int | None = None) -> List[List[Any]]:
    """
    Generate image match pairs from a list of view IDs. Assumes view_ids are sequential and
    wrap around. Each view is matched with the `window_radius` views before and after it.
    If `window_radius` is None or large enough to cover all views, exhaustive matching is used.

    Args:
        view_ids (List[str]): Ordered list of view identifiers.
        window_radius (Optional[int]): Number of neighbors to match on each side.

    Returns:
        rows (List[List[str]]): Each row is of the form [src, tgt1, tgt2, ...], meaning src
                                is matched to tgt1, tgt2, etc.
    """
    num_views = len(view_ids)

    if window_radius is None or 2*window_radius + 1 >= num_views:
        rows = [[view_ids[i] for i in range(j, num_views)] for j in range(num_views)]
    else:
        rows = []
        for i in range(num_views):
            rows.append([view_ids[i]])
            for j in range(window_radius):
                k =  i+j+1
                if k < num_views:
                    rows[i].append(view_ids[k])
                else:
                    rows[k % num_views].append(view_ids[i])
    rows = rows[:-1]
    return rows

def _make_pairs_sequential(view_ids: List[Any], window_radius: int) -> List[List[Any]]:
    """
    Generate image match pairs from a list of view IDs using a fixed sliding window.
    Each view is matched with the `window_radius` subsequent views, without wrapping
    around at the end of the list.

    Args:
        view_ids (List[Any]): Ordered list of view identifiers.
        window_radius (int): Number of forward neighbors to match for each view.

    Returns:
        List[List[Any]]: Each row is of the form [src, tgt1, tgt2, ...], meaning src
                         is matched to tgt1, tgt2, etc.
    """
    num_views = len(view_ids)
    rows = []
    for i in range(num_views-1):
        rows.append([view_ids[i]])
        for j in range(i+1, min(i+window_radius+1, num_views)):
            rows[i].append(view_ids[j])
    return rows

def _make_pairs_spatial(view_ids: List[Any], num_neighbors: int, locations: List[Tuple[float, float, float]]) -> List[List[Any]]:
    """
    Generate image match pairs from a list of view IDs based on spatial proximity.
    Each view is matched with its `num_neighbors` nearest neighbors using 3D Euclidean
    distance.

    Args:
        view_ids (List[Any]): List of view identifiers.
        num_neighbors (int): Number of nearest neighbors to match each view with.
        locations (List[Tuple[float, float, float]]): 3D coordinates corresponding to each view.

    Returns:
        List[List[Any]]: Each row is of the form [src, tgt1, tgt2, ...], where src is matched
                         to tgt1, tgt2, etc.
    """
    assert len(view_ids) == len(locations)
    num_views = len(view_ids)
    locations = np.array(locations)
    neighbors = [set() for _ in range(len(view_ids))]
    for i in range(len(view_ids)):
        dists = np.linalg.norm(locations - locations[i], axis=1)
        dists[i] = np.inf
        nearest = np.argsort(dists)[:min(num_neighbors, num_views-1)]
        for j in nearest:
            if j > i:
                neighbors[i].add(j)
            else:
                neighbors[j].add(i)

    #turn sets into list
    rows = []
    for i, neigh_set in enumerate(neighbors):
        if neigh_set:
            row = [view_ids[i]]
            for j in sorted(neigh_set):
                row.append(view_ids[j])
            rows.append(row)
    return rows


def write_pair_file(images: List[Path],
                    pairs: List[List[int]],
                    camera_init_file: Path,
                    output_path: Path) -> None:
    """
    Convert image index-based pairs to Meshroom internal view ID pairs using the provided
    camera_init_file. Write the result to a Meshroom compatible file.

    Args:
        images: A list of image Paths (ordered by index).
        pairs: A list of lists of indices into the `images` list.
        camera_init_file: Path to Meshroom's cameraInit.sfm or similar file containing view IDs.
        output_path: Path where the resulting pair file will be written.
    """
    path_to_viewid = _read_path_to_view_ids(camera_init_file)
    #convert the image id to Meshrooms view id
    rows = [[path_to_viewid[images[image_id]] for image_id in sublist] for sublist in pairs]
    with open(output_path, "w") as f:
        for row in rows:
            f.write(" ".join(map(str, row)) + "\n")


def _get_datetime_taken(image_path: Path) -> datetime.datetime | None:
    """Get the date and time from image metadata."""
    with Image.open(image_path) as img:
        value = img.getexif().get(306, None)
        if value is None:
            return None
        else:
            return datetime.datetime.strptime(value, '%Y:%m:%d %H:%M:%S')

def _extract_numbers(filename: str) -> List[int]:
    """Extracts all sub-strings of numbers as a list of ints."""
    return list(map(int, re.findall(r'\d+', filename)))

def preprocess_image_paths(paths: List[Path],
                           sort_by: str = 'filename',
                           sampling_rate: int = 1) -> List[Path]:
    """
    Sorts and sub-samples a list of image file paths.

    Args:
        paths (List[Path]): A list of image file paths to be processed.
        sort_by (str): Method to sort the images. Options are:
            - 'filename': Sort by numeric values extracted from filenames.
            - 'metadata': Sort by the DateTimeOriginal field in EXIF metadata.
        sampling_rate (int): Return every n-th image after sorting. Must be >= 1.

    Returns:
        List[Path]: A sorted and sub-sampled list of image paths.

    Raises:
        ValueError: If `sort_by` is not one of the valid options.
        ValueError: If `sampling_rate` is less than 1.
        ValueError: If sorting by metadata and no valid EXIF DateTimeOriginal is found.
    """
    if sampling_rate < 1:
        raise ValueError("sampling_rate must be a positive integer.")

    if sort_by == 'metadata':
        dated = []

        for path in paths:
            dt = _get_datetime_taken(path)
            if dt is None:
                raise ValueError(f"Tried to sort by metadata, but no valid 'DateTimeOriginal' metadata was found in image: {path}")
            dated.append((dt, path))

        dated.sort(key=lambda x: x[0])
        paths = [p for _, p in dated]
    elif sort_by == 'filename':
        paths = sorted(paths, key=lambda p: _extract_numbers(p.name))
    else:
        raise ValueError(f"Invalid sort_by value '{sort_by}'. Use 'metadata' or 'filename'.")

    return paths[::sampling_rate]

def get_meshroom_config_info(pipeline_path: str) -> Tuple[List[str], int]:
    """
    Extract the node names and downscale factor used in the DepthMap_1 node from a Meshroom pipeline.

    Args:
        pipeline_path (str): Path to the Meshroom pipeline JSON file.

    Returns:
        Tuple[List[str], int]: A list of node names in the pipeline, and the downscale factor.
            - Returns 2 if the downscale is not explicitly set (Meshroom default).
            - Returns -1 if the DepthMap_1 node is not present.
    """
    with open(pipeline_path) as f:
        pipeline_graph = json.load(f)['graph']

    nodes = list(pipeline_graph.keys())

    downsample_factor = (
        pipeline_graph['DepthMap_1']['inputs'].get('downscale', 2)
        if 'DepthMap_1' in pipeline_graph else -1
    )

    return nodes, downsample_factor



"""Helper utilities for unit tests"""
from __future__ import annotations

from dataclasses import fields, is_dataclass

import numpy as np
import xarray


def dataclasses_are_equal(obj1, obj2) -> bool:
    """Return whether two nested dataclass structures are equal by recursively checking for equality of fields,
    while specially handling numpy arrays and xarray Datasets.

    Recurses into dataclasses as well as dictionary-like, list-like, and tuple-like fields.
    """
    obj_type = type(obj1)
    if type(obj2) != obj_type:
        return False
    elif is_dataclass(obj_type):
        return all(
            dataclasses_are_equal(getattr(obj1, f.name), getattr(obj2, f.name))
            for f in fields(obj_type)
        )
    # handle the builtin types first for speed; subclasses handled below
    elif issubclass(obj_type, list) or issubclass(obj_type, tuple):
        return all(dataclasses_are_equal(v1,v2) for v1,v2 in zip(obj1,obj2))
    elif issubclass(obj_type, dict):
        if obj1.keys() != obj2.keys():
            return False
        return all(dataclasses_are_equal(obj1[k],obj2[k]) for k in obj1)
    elif issubclass(obj_type, np.ndarray):
        return (obj1==obj2).all()
    elif issubclass(obj_type, xarray.Dataset):
        return obj1.equals(obj2)
    else:
        return obj1 == obj2



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import Annotated, Literal, Tuple

import numpy as np
import pandas as pd
import xarray as xa

from openlifu.geo import Point
from openlifu.seg import SegmentationMethod
from openlifu.util.annotations import OpenLIFUFieldData
from openlifu.util.dict_conversion import DictMixin
from openlifu.util.units import getunitconversion, getunittype
from openlifu.xdc import Transducer

COORD_DIMS = ("x", "y", "z")
COORD_NAMES = ("Lateral", "Elevation", "Axial")

@dataclass
class SimSetup(DictMixin):

    spacing: Annotated[float, OpenLIFUFieldData("Spacing", "Simulation grid spacing")] = 1.0
    """Simulation grid spacing"""

    units: Annotated[str, OpenLIFUFieldData("Spatial units", "Units used for spatial measurements")] = "mm"
    """Units used for spatial measurements"""

    x_extent: Annotated[Tuple[float, float], OpenLIFUFieldData("X-extent", "Simulation grid extent along the first dimension")] = (-30., 30.)
    """Simulation grid extent along the first dimension"""

    y_extent: Annotated[Tuple[float, float], OpenLIFUFieldData("Y-extent", "Simulation grid extend along the second dimension")] = (-30., 30.)
    """Simulation grid extend along the second dimension"""

    z_extent: Annotated[Tuple[float, float], OpenLIFUFieldData("Z-extent", "Simulation grid extend along the third dimension")] = (-4., 60.)
    """Simulation grid extend along the third dimension"""

    dt: Annotated[float, OpenLIFUFieldData("Time step", "Simulation time step")] = 0.
    """Simulation time step"""

    t_end: Annotated[float, OpenLIFUFieldData("End time", """Simulation end time""")] = 0.
    """Simulation end time"""

    c0: Annotated[float, OpenLIFUFieldData("Speed of Sound (m/s)", "Reference speed of sound for converting distance to time")] = 1500.0
    """Reference speed of sound for converting distance to time"""

    cfl: Annotated[float, OpenLIFUFieldData("CFL number", "Courant-Friedrichs-Lewy number")] = 0.5
    """Courant-Friedrichs-Lewy number"""

    options: Annotated[dict[str, str], OpenLIFUFieldData("Simulation options", "Additional simulation options")] = field(default_factory=dict)
    """Additional simulation options"""

    def __post_init__(self):
        if len(self.x_extent) != 2:
            raise ValueError("x_extent must have length 2.")
        if self.x_extent[0] >= self.x_extent[1]:
            raise ValueError("x_extent must be in the form (min, max) with min < max.")
        if len(self.y_extent) != 2:
            raise ValueError("y_extent must have length 2.")
        if self.y_extent[0] >= self.y_extent[1]:
            raise ValueError("y_extent must be in the form (min, max) with min < max.")
        if len(self.z_extent) != 2:
            raise ValueError("z_extent must have length 2.")
        if self.z_extent[0] >= self.z_extent[1]:
            raise ValueError("z_extent must be in the form (min, max) with min < max.")
        if not isinstance(self.spacing, int | float):
            raise TypeError("spacing must be a number.")
        if self.spacing <= 0:
            raise ValueError("spacing must be a positive number.")
        if not isinstance(self.units, str):
            raise TypeError("units must be a string.")
        if getunittype(self.units) != 'distance':
            raise ValueError(f"units must be a length unit, got {self.units}.")
        if not isinstance(self.c0, int | float):
            raise TypeError("c0 must be a number.")
        if self.c0 <= 0:
            raise ValueError("c0 must be a positive number.")
        if not isinstance(self.cfl, int | float):
            raise TypeError("cfl must be a number.")
        if self.cfl <= 0:
            raise ValueError("cfl must be a positive number.")
        if not isinstance(self.dt, int | float):
            raise TypeError("dt must be a number.")
        if self.dt < 0:
            raise ValueError("dt must be a non-negative number.")
        if not isinstance(self.t_end, int | float):
            raise TypeError("t_end must be a number.")
        if self.t_end < 0:
            raise ValueError("t_end must be a non-negative number.")
        nx = np.diff(self.x_extent)/self.spacing
        x_extent = tuple(np.arange(2)*np.round(nx)*self.spacing + self.x_extent[0])
        if ((0.5-np.abs((nx % 1) - 0.5))/ np.round(nx)) > 1e-3:
            logging.warning(f"x_extent {self.x_extent} does not evenly divide by spacing ({self.spacing}). Rounding to {x_extent}.")
        self.x_extent = x_extent
        ny = np.diff(self.y_extent)/self.spacing
        y_extent = tuple(np.arange(2)*np.round(ny)*self.spacing + self.y_extent[0])
        if ((0.5-np.abs((ny % 1) - 0.5))/ np.round(ny)) > 1e-3:
            logging.warning(f"y_extent {self.y_extent} does not evenly divide by spacing ({self.spacing}). Rounding to {y_extent}.")
        self.y_extent = y_extent
        nz = np.diff(self.z_extent)/self.spacing
        z_extent = tuple(np.arange(2)*np.round(nz)*self.spacing + self.z_extent[0])
        if ((0.5-np.abs((nz % 1) - 0.5))/ np.round(nz)) > 1e-3:
            logging.warning(f"z_extent {self.z_extent} does not evenly divide by spacing ({self.spacing}). Rounding to {z_extent}.")
        self.z_extent = z_extent

    def get_coords(self, dims=None, units: str | None = None):
        dims = COORD_DIMS if dims is None else dims
        units = self.units if units is None else units
        sizes = self.get_size(dims)
        extents = self.get_extent(dims, units)
        coords = xa.Coordinates({dim: np.linspace(extents[i][0], extents[i][1], sizes[i]) for i, dim in enumerate(dims)})
        for dim in dims:
            coords[dim].attrs['units'] = units
            coords[dim].attrs['long_name'] = COORD_NAMES[COORD_DIMS.index(dim)]
        return coords

    def get_corners(self, units: str | None = None):
        units = self.units if units is None else units
        scl = getunitconversion(self.units, units)
        xyz = np.array(np.meshgrid(self.x_extent, self.y_extent, self.z_extent, indexing='ij'))
        corners = xyz.reshape(3,-1)
        return corners*scl

    def get_extent(self, dims: str | None=None, units: str | None = None):
        dims = COORD_DIMS if dims is None else dims
        units = self.units if units is None else units
        scl = getunitconversion(self.units, units)
        extents = [self.x_extent, self.y_extent, self.z_extent]
        return np.array([extents[COORD_DIMS.index(dim)] for dim in dims])*scl

    def get_max_cycle_offset(self, arr:Transducer, frequency: float | None = None, delays: np.ndarray | None=None, zmin: float =10e-3):
        frequency = arr.frequency if frequency is None else frequency
        delays = np.zeros(arr.numelements()) if delays is None else delays
        coords = self.get_coords(units="m")
        dims = coords.dims
        cvals = [coords[dims[0]], coords[dims[1]], coords[dims[2]].sel(ax=slice(zmin, None))]
        ndg = np.meshgrid(*cvals)
        dists = [np.sqrt((ndg[0]-pos[0])**2 + (ndg[1]-pos[1])**2 + (ndg[2]-pos[2])**2) for pos in arr.get_positions(units="m")]
        tof = [dist/self.c0 + delays[i] for i, dist in enumerate(dists)]
        dtof = np.array(tof).max(axis=0) - np.array(tof).min(axis=0)
        max_cycle_offset = dtof.max()*frequency
        return max_cycle_offset

    def get_max_distance(self, arr: Transducer, units: str | None = None):
        units = self.units if units is None else units
        corners = self.get_corners(units=units)
        distances = np.array([[el.distance_to_point(corner, units=units) for corner in corners.T] for el in arr.elements])
        max_distance = np.max(distances)
        return max_distance

    def get_size(self, dims: str | None=None):
        dims = COORD_DIMS if dims is None else dims
        n = [int((np.round(np.diff(ext)/self.spacing)).item())+1 for ext in [self.x_extent, self.y_extent, self.z_extent]]
        return np.array([n[COORD_DIMS.index(dim)] for dim in dims]).squeeze()

    def get_spacing(self, units: str | None = None):
        units = self.units if units is None else units
        return getunitconversion(self.units, units)*self.spacing

    def setup_sim_scene(
            self,
            seg_method: SegmentationMethod,
            volume: xa.DataArray | None = None
        ) -> Tuple[xa.DataArray, Transducer, Point]:
        """ Prepare a simulation scene composed of a simulation grid

        Setup a simulation scene with a simulation grid including physical properties.
        A segmentation is performed to detect the medium, so we can assign
        physical properties to each voxel, later used by the ultrasound simulation.
        This assume that the input volume is resampled to the geo-referenced simulation grid (lon, lat, ele).

        Args:
            seg_method: seg.SegmentationMethod
            volume: xa.DataArray
                Optional volume to be used for simulation grid definition (Default: None).
                The volume is assumed to be resampled on sim grid coordinates.

        Returns
            params: The xa.DataArray simulation grid with physical properties for each voxel
        """
        if volume is None:
            sim_coords = self.get_coords()
            params = seg_method.ref_params(sim_coords)
        else:
            params = seg_method.seg_params(volume)

        return params

    def to_table(self) -> pd.DataFrame:
        """
        Get a table of the simulation setup parameters

        :returns: Pandas DataFrame of the simulation setup parameters
        """
        records = [
            {"Name": "Spacing", "Value": self.spacing, "Unit": self.units},
            {"Name": "X Extent", "Value": f"{self.x_extent[0]} to {self.x_extent[1]}", "Unit": self.units},
            {"Name": "Y Extent", "Value": f"{self.y_extent[0]} to {self.y_extent[1]}", "Unit": self.units},
            {"Name": "Z Extent", "Value": f"{self.z_extent[0]} to {self.z_extent[1]}", "Unit": self.units},
            {"Name": "Time Step", "Value": self.dt, "Unit": "s"},
            {"Name": "End Time", "Value": self.t_end, "Unit": "s"},
            {"Name": "Speed of Sound", "Value": self.c0, "Unit": "m/s"},
            {"Name": "CFL", "Value": self.cfl, "Unit": ""},
        ]
        return pd.DataFrame.from_records(records)

    @staticmethod
    def from_dict(d: dict, on_keyword_mismatch: Literal['warn', 'raise', 'ignore'] = 'warn') -> SimSetup:
        """Create a SimSetup instance from a dictionary."""
        if not isinstance(d, dict):
            raise TypeError("Input must be a dictionary.")

        expected_keywords = [
            'spacing', 'units', 'x_extent', 'y_extent', 'z_extent',
            'dt', 't_end', 'c0', 'cfl', 'options'
        ]

        input_args = {
            k: v for k, v in d.items() if k in expected_keywords
        }
        unexpected_keywords = [k for k in d if k not in expected_keywords]

        if unexpected_keywords:
            if on_keyword_mismatch == 'raise':
                raise TypeError(f"Unexpected keyword arguments for SimSetup: {unexpected_keywords}")
            elif on_keyword_mismatch == 'warn':
                logging.warning(f"Ignoring unexpected keyword arguments for SimSetup: {unexpected_keywords}")

        return SimSetup(**input_args)



from __future__ import annotations

from dataclasses import fields
from datetime import datetime, timedelta
from pathlib import Path

import numpy as np
import pytest
import xarray as xa
from helpers import dataclasses_are_equal

from openlifu import Point, Pulse, Sequence, Solution, Transducer
from openlifu.bf.focal_patterns import SinglePoint
from openlifu.plan import SolutionAnalysis
from openlifu.xdc.element import Element


@pytest.fixture()
def example_transducer() -> Transducer:
    return Transducer(
        id="trans_456",
        name="Test Transducer",
        elements=[
            Element(index=1, position=[-14, -14, 0], units="mm"),
            Element(index=2, position=[-2, -2, 0], units="mm"),
            Element(index=3, position=[2, 2, 0], units="mm"),
            Element(index=4, position=[14, 14, 0], units="mm")
        ],
        frequency=1e6,
        units="mm"
    )


@pytest.fixture()
def example_focal_pattern_single() -> SinglePoint:
    return SinglePoint(target_pressure=1.0e6, units="Pa")


@pytest.fixture()
def example_solution() -> Solution:
    rng = np.random.default_rng(147)
    return Solution(
        id="sol_001",
        name="Test Solution",
        protocol_id="prot_123",
        transducer=Transducer(
            id="trans_456",
            name="Test Transducer",
            elements=[
                Element(index=1, position=[-14, -14, 0], units="m"),
                Element(index=2, position=[-2, -2, 0], units="m"),
                Element(index=3, position=[2, 2, 0], units="m"),
                Element(index=4, position=[14, 14, 0], units="m")
            ],
            frequency=1e6,
            units="m"
        ),
        date_created=datetime(2024, 1, 1, 12, 0),
        description="This is a test solution for a unit test.",
        delays=np.array([[0.0, 1.0, 2.0, 3.0]]),
        apodizations=np.array([[0.5, 0.75, 1.0, 0.85]]),
        pulse=Pulse(frequency=42),
        sequence=Sequence(pulse_count=27, pulse_interval=2, pulse_train_interval=2*27+5),
        foci=[Point(id="test_focus_point")],
        target=Point(id="test_target_point"),
        simulation_result=xa.Dataset(
            {
                'p_min': xa.DataArray(
                    data=rng.random((1, 3, 2, 3)),
                    dims=["focal_point_index", "x", "y", "z"],
                    attrs={'units': "Pa"}
                ),
                'p_max': xa.DataArray(
                    data=rng.random((1, 3, 2, 3)),
                    dims=["focal_point_index", "x", "y", "z"],
                    attrs={'units': "Pa"}
                ),
                'intensity': xa.DataArray(
                    data=rng.random((1, 3, 2, 3)),
                    dims=["focal_point_index", "x", "y", "z"],
                    attrs={'units': "W/cm^2"}
                )
            },
            coords={
                'x': xa.DataArray(dims=["x"], data=np.linspace(0, 1, 3), attrs={'units': "m"}),
                'y': xa.DataArray(dims=["y"], data=np.linspace(0, 1, 2), attrs={'units': "m"}),
                'z': xa.DataArray(dims=["z"], data=np.linspace(0, 1, 3), attrs={'units': "m"}),
                'focal_point_index': [0]
            }
        )
    )


def test_default_solution():
    """Ensure it is possible to construct a default Solution"""
    Solution()


@pytest.mark.parametrize("compact_representation", [True, False])
@pytest.mark.parametrize("include_simulation_data", [True, False])
@pytest.mark.parametrize("default_solution", [True, False])
def test_json_serialize_deserialize_solution(
    example_solution: Solution,
    compact_representation: bool,
    include_simulation_data: bool,
    default_solution: bool
):
    """Verify that turning a Solution into json and then re-constructing it gets back to the original solution"""

    # Default solution serializes a bit differently because it's full of None values, so we test both cases
    solution = Solution() if default_solution else example_solution
    solution_json = solution.to_json(include_simulation_data=include_simulation_data, compact=compact_representation)
    if include_simulation_data:
        solution_reconstructed = Solution.from_json(solution_json)
    else:
        solution_reconstructed = Solution.from_json(solution_json, simulation_result=solution.simulation_result)
    assert dataclasses_are_equal(solution_reconstructed, solution)


def test_save_load_solution(example_solution: Solution, tmp_path: Path):
    """Test that a solution can be saved to and loaded from disk faithfully."""
    json_filepath = tmp_path/"some_directory"/"example_soln.json"
    example_solution.to_files(json_filepath)
    assert dataclasses_are_equal(Solution.from_files(json_filepath), example_solution)


def test_save_load_solution_custom_dataset_filepath(example_solution: Solution, tmp_path: Path):
    """Test that a solution can be saved to and loaded from disk faithfully, this time with a custom path for simulation data."""
    json_filepath = tmp_path/"some_directory"/"example_soln.json"
    nc_filepath = tmp_path/"some_other_directory"/"sim_output.nc"
    example_solution.to_files(json_filepath, nc_filepath)
    assert dataclasses_are_equal(Solution.from_files(json_filepath, nc_filepath), example_solution)


def test_num_foci(example_solution:Solution):
    """Ensure that the number of foci in the test solution matches the number of foci provided in the simuluation and beamform data."""
    num_foci = example_solution.num_foci()
    assert len(example_solution.foci) == num_foci
    assert len(example_solution.simulation_result['focal_point_index']) == num_foci
    assert example_solution.delays.shape[0] == num_foci
    assert example_solution.apodizations.shape[0] == num_foci

@pytest.mark.parametrize("compact_representation", [True, False])
def test_json_serialize_deserialize_solution_analysis(compact_representation: bool):
    """Verify that turning a SolutionAnalysis into json and then re-constructing it gets back to the original"""
    analysis = SolutionAnalysis(mainlobe_isppa_Wcm2=[1,2],beamwidth_ax_6dB_mm=[3,4], MI=5)
    analysis_json = analysis.to_json(compact=compact_representation)
    analysis_reconstructed = SolutionAnalysis.from_json(analysis_json)
    assert dataclasses_are_equal(analysis_reconstructed, analysis)

def test_solution_analyze_data_types(example_solution:Solution):
    """Test that solution analysis field are all floats or lists of floats as expected"""
    analysis = example_solution.analyze()
    for f in fields(analysis):
        value = getattr(analysis, f.name)
        if f.name == "param_constraints":
            assert isinstance(value, dict)
        elif not isinstance(value, float):
                assert isinstance(value, list)
                if len(value) > 0:
                    assert isinstance(value[0], float)


def test_solution_created_date():
    """Test that created date is recent when a solution is created."""
    tolerance = timedelta(seconds=2)

    solution = Solution()
    now = datetime.now()
    assert(now - tolerance <= solution.date_created <= now + tolerance)


def test_solution_analyze_ratios(example_solution: Solution):
    """Test the calculation of mainlobe to sidelobe ratios in Solution.analyze()"""
    solution = example_solution
    # Use only one focus point for simplicity
    solution.foci = [Point(id="test_focus_point", position=np.array([0, 0, 0.05]), units="m")] # Centered focus
    solution.delays = np.array([[0.0, 0.0, 0.0, 0.0]])
    solution.apodizations = np.array([[1.0, 1.0, 1.0, 1.0]])

    # Create a simple simulation result
    # Dimensions: focal_point_index, lat, ele, ax
    # Coords: lat=[-0.01, 0, 0.01], ele=[0], ax=[0.04, 0.05, 0.06] (units in m)
    lat_coords = np.array([-0.01, 0, 0.01])
    ele_coords = np.array([0])
    ax_coords = np.array([0.04, 0.05, 0.06])

    p_min_data = np.zeros((1, 3, 1, 3))  # Initialize with zeros
    intensity_data = np.zeros((1, 3, 1, 3)) # Initialize with zeros

    # Case 1: Normal values
    # Mainlobe at (lat=0, ele=0, ax=0.05)
    # Sidelobe at (lat=0.01, ele=0, ax=0.06) -> make this distinct
    p_min_data[0, 1, 0, 1] = 1.0e6  # Mainlobe pressure = 1 MPa
    p_min_data[0, 2, 0, 2] = 0.5e6  # Sidelobe pressure = 0.5 MPa
    intensity_data[0, 1, 0, 1] = 10.0  # Mainlobe intensity = 10 W/cm^2
    intensity_data[0, 2, 0, 2] = 2.0   # Sidelobe intensity = 2 W/cm^2

    solution.simulation_result = xa.Dataset(
        {
            'p_min': xa.DataArray(data=p_min_data, dims=["focal_point_index", "x", "y", "z"], attrs={'units': "Pa"}),
            'p_max': xa.DataArray(data=p_min_data, dims=["focal_point_index", "x", "y", "z"], attrs={'units': "Pa"}), # Keep p_max same for simplicity
            'intensity': xa.DataArray(data=intensity_data, dims=["focal_point_index", "x", "y", "z"], attrs={'units': "W/cm^2"})
        },
        coords={
            'x': xa.DataArray(dims=["x"], data=lat_coords, attrs={'units': "m"}),
            'y': xa.DataArray(dims=["y"], data=ele_coords, attrs={'units': "m"}),
            'z': xa.DataArray(dims=["z"], data=ax_coords, attrs={'units': "m"}),
            'focal_point_index': [0]
        }
    )

    from openlifu.plan.solution_analysis import SolutionAnalysisOptions
    # Ensure options make the chosen voxels the max in their respective regions
    # Focus is at (0,0,0.05). Mainlobe radius will capture p_min_data[0,1,0,1]
    # Sidelobe radius will capture p_min_data[0,2,0,2] if it's outside mainlobe.
    options = SolutionAnalysisOptions(
        mainlobe_radius=0.005, # 5mm, should capture central voxel
        sidelobe_radius=0.005, # 5mm, but mask is > sidelobe_radius, so it's outside this
                                 # and aspect ratio matters.
        mainlobe_aspect_ratio=(1,1,1), # Make it spherical for simplicity
        sidelobe_zmin=0.001, # well below focus
        distance_units="m" # Matching simulation_result
    )
    # The transducer effective origin also plays a role in get_mask.
    # For example_transducer, with all elements at Z=0 and centered apodization, origin is approx (0,0,0)
    # Let's assume default transducer origin is [0,0,0] for get_mask calculations.

    analysis = solution.analyze(options=options)

    # Check Case 1
    # Expected mainlobe pnp = 1.0 MPa (from p_min_data[0,1,0,1] = 1e6 Pa)
    # Expected sidelobe pnp = 0.5 MPa (from p_min_data[0,2,0,2] = 0.5e6 Pa)
    # Expected pressure ratio = 0.5 / 1.0 = 0.5

    # Expected mainlobe isppa = 10 W/cm^2
    # Expected sidelobe isppa = 2 W/cm^2
    # Expected intensity ratio = 2.0 / 10.0 = 0.2

    # Due to potential small variations from interpolation/masking, use approx
    assert np.isclose(analysis.mainlobe_pnp_MPa[0], 1.0), f"Mainlobe PNP was {analysis.mainlobe_pnp_MPa[0]}"
    assert analysis.sidelobe_pnp_MPa[0] > 0, "Sidelobe PNP was zero, check mask/data for Case 1." # Ensure it's not zero
    # Using the actual found sidelobe value for ratio assertion robustness
    expected_pressure_ratio_case1 = analysis.sidelobe_pnp_MPa[0] / analysis.mainlobe_pnp_MPa[0]
    assert np.isclose(analysis.sidelobe_to_mainlobe_pressure_ratio[0], expected_pressure_ratio_case1), \
        f"Pressure ratio calculation error. Got {analysis.sidelobe_to_mainlobe_pressure_ratio[0]}, expected {expected_pressure_ratio_case1}"

    assert np.isclose(analysis.mainlobe_isppa_Wcm2[0], 10.0), f"Mainlobe ISPPA was {analysis.mainlobe_isppa_Wcm2[0]}"
    assert analysis.sidelobe_isppa_Wcm2[0] > 0, "Sidelobe ISPPA was zero, check mask/data for Case 1." # Ensure it's not zero
    expected_intensity_ratio_case1 = analysis.sidelobe_isppa_Wcm2[0] / analysis.mainlobe_isppa_Wcm2[0]
    assert np.isclose(analysis.sidelobe_to_mainlobe_intensity_ratio[0], expected_intensity_ratio_case1), \
        f"Intensity ratio calculation error. Got {analysis.sidelobe_to_mainlobe_intensity_ratio[0]}, expected {expected_intensity_ratio_case1}"


    # Case 2: Sidelobe pressure is zero, Mainlobe is non-zero
    p_min_data_case2 = np.zeros((1, 3, 1, 3))
    p_min_data_case2[0, 1, 0, 1] = 1.0e6  # Mainlobe pressure = 1 MPa
    p_min_data_case2[0, 2, 0, 2] = 0.0    # Sidelobe pressure = 0 MPa
    solution.simulation_result['p_min'].data = p_min_data_case2
    solution.simulation_result['intensity'].data = intensity_data # Keep intensity same

    analysis_case2 = solution.analyze(options=options)
    assert analysis_case2.mainlobe_pnp_MPa[0] == 1.0
    assert analysis_case2.sidelobe_pnp_MPa[0] == 0.0
    assert analysis_case2.sidelobe_to_mainlobe_pressure_ratio[0] == 0.0 # 0 / 1.0 == 0.0

    # Case 3: Sidelobe intensity is zero, Mainlobe is non-zero
    intensity_data_case3 = np.zeros((1,3,1,3))
    intensity_data_case3[0, 1, 0, 1] = 10.0  # Mainlobe intensity = 10 W/cm^2
    intensity_data_case3[0, 2, 0, 2] = 0.0   # Sidelobe intensity = 0 W/cm^2
    solution.simulation_result['p_min'].data = p_min_data # Reset p_min
    solution.simulation_result['intensity'].data = intensity_data_case3

    analysis_case3 = solution.analyze(options=options)
    assert analysis_case3.mainlobe_isppa_Wcm2[0] == 10.0
    assert analysis_case3.sidelobe_isppa_Wcm2[0] == 0.0
    assert analysis_case3.sidelobe_to_mainlobe_intensity_ratio[0] == 0.0 # 0 / 10.0 == 0.0

    # Case 4: Mainlobe pressure is zero, Sidelobe is non-zero
    p_min_data_case4 = np.zeros((1, 3, 1, 3))
    p_min_data_case4[0, 1, 0, 1] = 0.0    # Mainlobe pressure = 0 MPa
    p_min_data_case4[0, 2, 0, 2] = 0.5e6  # Sidelobe pressure = 0.5 MPa
    solution.simulation_result['p_min'].data = p_min_data_case4
    solution.simulation_result['intensity'].data = intensity_data # Keep intensity same

    analysis_case4 = solution.analyze(options=options)
    assert analysis_case4.mainlobe_pnp_MPa[0] == 0.0
    # Ensure sidelobe is picked up as non-zero
    assert analysis_case4.sidelobe_pnp_MPa[0] > 0, "Sidelobe PNP was zero for Case 4, expected non-zero."
    assert analysis_case4.sidelobe_to_mainlobe_pressure_ratio[0] == np.inf # 0.5 / 0 == inf

    # Case 5: Mainlobe intensity is zero, Sidelobe is non-zero
    intensity_data_case5 = np.zeros((1,3,1,3))
    intensity_data_case5[0, 1, 0, 1] = 0.0   # Mainlobe intensity = 0 W/cm^2
    intensity_data_case5[0, 2, 0, 2] = 2.0  # Sidelobe intensity = 2 W/cm^2
    solution.simulation_result['p_min'].data = p_min_data # Reset p_min
    solution.simulation_result['intensity'].data = intensity_data_case5
    analysis_case5 = solution.analyze(options=options)
    assert analysis_case5.mainlobe_isppa_Wcm2[0] == 0.0
    assert analysis_case5.sidelobe_isppa_Wcm2[0] > 0, "Sidelobe ISPPA was zero for Case 5, expected non-zero."
    assert analysis_case5.sidelobe_to_mainlobe_intensity_ratio[0] == np.inf # 2.0 / 0 == inf

    # Case 6: Mainlobe and Sidelobe pressure are zero
    p_min_data_case6 = np.zeros((1, 3, 1, 3))
    solution.simulation_result['p_min'].data = p_min_data_case6
    solution.simulation_result['intensity'].data = intensity_data # Keep intensity same

    analysis_case6 = solution.analyze(options=options)
    assert analysis_case6.mainlobe_pnp_MPa[0] == 0.0
    assert analysis_case6.sidelobe_pnp_MPa[0] == 0.0
    assert np.isnan(analysis_case6.sidelobe_to_mainlobe_pressure_ratio[0]) # 0 / 0 == nan

    # Case 7: Mainlobe and Sidelobe intensity are zero
    intensity_data_case7 = np.zeros((1,3,1,3))
    solution.simulation_result['p_min'].data = p_min_data # Reset p_min
    solution.simulation_result['intensity'].data = intensity_data_case7
    analysis_case7 = solution.analyze(options=options)
    assert analysis_case7.mainlobe_isppa_Wcm2[0] == 0.0
    assert analysis_case7.sidelobe_isppa_Wcm2[0] == 0.0
    assert np.isnan(analysis_case7.sidelobe_to_mainlobe_intensity_ratio[0]) # 0 / 0 == nan



from __future__ import annotations

import json

from openlifu.util.types import PathLike
from openlifu.xdc.transducer import Transducer
from openlifu.xdc.transducerarray import TransducerArray


def load_transducer_from_file(transducer_filepath : PathLike, convert_array:bool = True) -> Transducer|TransducerArray:
    """Load a Transducer or TransducerArray from file, depending on the "type" field in the file.
    Note: the transducer object includes the relative path to the affiliated transducer model data. `get_transducer_absolute_filepaths`, should
    be used to obtain the absolute data filepaths based on the Database directory path.
    Args:
        transducer_filepath: path to the transducer json file
        convert_array: When enabled, if a TransducerArray is encountered then it is converted to a Transducer.
    Returns: a Transducer if the json file defines a Transducer, or if the json file defines a TransducerArray and convert_array is enabled.
        Otherwise a TransducerArray.
    """
    with open(transducer_filepath) as f:
        if not f:
            raise FileNotFoundError(f"Transducer file not found at: {transducer_filepath}")
        d = json.load(f)
    if "type"  in d and d["type"] == "TransducerArray":
        transducer = TransducerArray.from_dict(d)
        if convert_array:
            transducer = transducer.to_transducer()
    else:
        transducer = Transducer.from_file(transducer_filepath)
    return transducer



from __future__ import annotations

import logging
from pathlib import Path

import numpy as np
import pytest
import xarray as xa
from pytest_mock import MockerFixture

from openlifu import Point, Protocol, Transducer
from openlifu.bf.focal_patterns import Wheel
from openlifu.db import Session
from openlifu.plan.protocol import OnPulseMismatchAction
from openlifu.plan.target_constraints import TargetConstraints


@pytest.fixture()
def example_protocol() -> Protocol:
    return Protocol.from_file(Path(__file__).parent/'resources/example_db/protocols/example_protocol/example_protocol.json')

@pytest.fixture()
def example_transducer() -> Transducer:
    return Transducer.from_file(Path(__file__).parent/"resources/example_db/transducers/example_transducer/example_transducer.json")

@pytest.fixture()
def example_session() -> Session:
    return Session.from_file(Path(__file__).parent/"resources/example_db/subjects/example_subject/sessions/example_session/example_session.json")

@pytest.fixture()
def example_wheel_pattern() -> Wheel:
    return Wheel(num_spokes=6)

def test_to_dict_from_dict(example_protocol: Protocol):
    proto_dict = example_protocol.to_dict()
    new_protocol = Protocol.from_dict(proto_dict)
    assert new_protocol == example_protocol

@pytest.mark.parametrize("compact_representation", [True, False])
def test_serialize_deserialize_protocol(example_protocol : Protocol, compact_representation: bool):
    assert example_protocol.from_json(example_protocol.to_json(compact_representation)) == example_protocol

def test_default_protocol():
    """Ensure it is possible to construct a default protocol"""
    Protocol()

def test_to_table(example_protocol: Protocol):
    """Ensure that the protocol can be correctly converted to a table."""
    t = example_protocol.to_table()
    assert t is not None
    assert "Category" in t.columns
    assert "Name" in t.columns
    assert "Value" in t.columns
    assert "Unit" in t.columns
    tm =t.set_index(["Category", "Name"])
    assert tm.loc["","ID"]["Value"] == example_protocol.id
    assert tm.loc["Delay Method", "Default Sound Speed"]["Value"] == 1540.0
    assert tm.loc["Delay Method", "Default Sound Speed"]["Unit"] == "m/s"

@pytest.mark.parametrize(
    "target_constraints",
    [
        [
            TargetConstraints(dim="P", units="mm", min=0.0, max=float("inf")),
        ],
        [
            TargetConstraints(dim="P", units="m", min=-0.001, max=0.0),
        ],
        [
            TargetConstraints(dim="L", units="mm", min=-100.0, max=0.0),
            TargetConstraints(dim="P", units="mm", min=-100.0, max=0.0),
            TargetConstraints(dim="S", units="mm", min=-100.0, max=-10.0),
        ]
    ]
)
def test_check_target(example_protocol: Protocol, example_session: Session, target_constraints: TargetConstraints):
    """Ensure that the target can be correctly verified."""
    example_protocol.target_constraints = target_constraints
    with pytest.raises(ValueError, match="not within bounds"):
        example_protocol.check_target(example_session.targets[0])

@pytest.mark.parametrize("on_pulse_mismatch", [
            OnPulseMismatchAction.ERROR,
            OnPulseMismatchAction.ROUND,
            OnPulseMismatchAction.ROUNDUP,
            OnPulseMismatchAction.ROUNDDOWN
        ]
    )
def test_fix_pulse_mismatch(
        example_protocol: Protocol,
        example_session: Session,
        example_wheel_pattern: Wheel,
        on_pulse_mismatch: OnPulseMismatchAction
    ):
    """Test if sequence is correctly fixed for all pulse mismatch actions."""
    logging.disable(logging.CRITICAL)

    target = example_session.targets[0]
    foci = example_wheel_pattern.get_targets(target)
    num_foci = len(foci)
    if on_pulse_mismatch is OnPulseMismatchAction.ERROR:
        with pytest.raises(ValueError, match="not a multiple of the number of foci"):
            example_protocol.fix_pulse_mismatch(on_pulse_mismatch, foci)
    else:
        example_protocol.fix_pulse_mismatch(on_pulse_mismatch, foci)
        if on_pulse_mismatch is OnPulseMismatchAction.ROUND:
            assert example_protocol.sequence.pulse_count == num_foci
        elif on_pulse_mismatch is OnPulseMismatchAction.ROUNDUP:
            assert example_protocol.sequence.pulse_count == 2*num_foci
        elif on_pulse_mismatch is OnPulseMismatchAction.ROUNDDOWN:
            assert example_protocol.sequence.pulse_count == num_foci

@pytest.mark.parametrize("use_gpu", [True, False, None])
@pytest.mark.parametrize("gpu_is_available", [True, False])
def test_calc_solution_use_gpu(
    mocker:MockerFixture,
    example_protocol:Protocol,
    example_transducer:Transducer,
    use_gpu:bool | None,
    gpu_is_available:bool,
):
    """Test that the correct value of use_gpu is passed to the simulation runner"""
    example_simulation_output = xa.Dataset(
        {
            'p_min': xa.DataArray(data=np.empty((3, 2, 3)), dims=["x", "y", "z"], attrs={'units': "Pa"}),
            'p_max': xa.DataArray(data=np.empty((3, 2, 3)),dims=["x", "y", "z"],attrs={'units': "Pa"}),
            'intensity': xa.DataArray(data=np.empty((3, 2, 3)),dims=["x", "y", "z"],attrs={'units': "W/cm^2"}),
        },
        coords={
            'x': xa.DataArray(dims=["x"], data=np.linspace(0, 1, 3), attrs={'units': "m"}),
            'y': xa.DataArray(dims=["y"], data=np.linspace(0, 1, 2), attrs={'units': "m"}),
            'z': xa.DataArray(dims=["z"], data=np.linspace(0, 1, 3), attrs={'units': "m"}),
        },
    )
    mocker.patch(
        "openlifu.plan.protocol.gpu_available",
        return_value = gpu_is_available,
    )
    run_simulation_mock = mocker.patch(
        "openlifu.plan.protocol.run_simulation",
        return_value = (example_simulation_output, None),
    )
    example_protocol.calc_solution(
        target = Point(),
        transducer = example_transducer,
        simulate = True,
        scale = False,
        use_gpu=use_gpu,
    )
    args, kwargs = run_simulation_mock.call_args
    if use_gpu is None:
        assert kwargs['gpu'] == gpu_is_available
    else:
        assert kwargs['gpu'] == use_gpu



from __future__ import annotations

import importlib.resources
import shutil
from pathlib import Path

import numpy as np
import pytest
from vtk import VTK_UNSIGNED_SHORT, vtkImageData, vtkPoints, vtkPolyData

from openlifu.db.database import Database
from openlifu.nav.photoscan import (
    Photoscan,
    _make_pairs_sequential,
    _make_pairs_sequential_loop,
    _make_pairs_spatial,
    apply_exif_orientation_numpy,
    convert_between_ras_and_lps,
    convert_numpy_to_vtkimage,
    get_meshroom_pipeline_names,
    load_data_from_filepaths,
    load_data_from_photoscan,
    preprocess_image_modnet,
)


@pytest.fixture()
def example_database(tmp_path:Path) -> Database:
    """Example database in a temporary directory"""
    shutil.copytree(Path(__file__).parent/'resources/example_db', tmp_path/"example_db")
    return Database(tmp_path/"example_db")

@pytest.fixture()
def example_photoscan() -> Photoscan:
    return Photoscan.from_file(Path(__file__).parent/'resources/example_db/subjects/example_subject/sessions/example_session/photoscans/example_photoscan/example_photoscan.json')

@pytest.mark.parametrize("compact_representation", [True, False])
def test_serialize_deserialize_photoscan(example_photoscan : Photoscan, compact_representation: bool):
    reconstructed_photoscan =  example_photoscan.from_json(example_photoscan.to_json(compact_representation))
    assert example_photoscan == reconstructed_photoscan

def test_load_photoscan(example_database:Database, tmp_path:Path):

    subject_id = "example_subject"
    session_id = "example_session"
    photoscan_id = "example_photoscan"
    photoscan_info = example_database.get_photoscan_absolute_filepaths_info(subject_id, session_id, photoscan_id)
    [model, texture] = load_data_from_filepaths(photoscan_info["model_abspath"], photoscan_info["texture_abspath"])
    assert model is not None
    assert texture is not None
    assert isinstance(model, vtkPolyData)
    assert isinstance(texture,vtkImageData)

    # From photoscan object
    photoscan_metadata_filepath = example_database.get_photoscan_metadata_filepath(subject_id, session_id, photoscan_id)
    photoscan = Photoscan.from_file(photoscan_metadata_filepath)
    [model, texture] = load_data_from_photoscan(photoscan, parent_dir = Path(photoscan_metadata_filepath).parent)
    assert model is not None
    assert texture is not None
    assert isinstance(model, vtkPolyData)
    assert isinstance(texture,vtkImageData)

    bogus_file = Path(tmp_path/"test_db_files/bogus_photoscan.pdf")
    bogus_file.parent.mkdir(parents=True, exist_ok=True)
    bogus_file.touch()
    with pytest.raises(ValueError, match="not supported by reader"):
        load_data_from_filepaths(photoscan_info["model_abspath"], bogus_file)
    with pytest.raises(ValueError, match="not supported by reader"):
        load_data_from_filepaths(bogus_file,photoscan_info["texture_abspath"])

    # File does not exist
    bogus_file = Path(tmp_path/"test_db_files/bogus_photoscan.obj")
    with pytest.raises(FileNotFoundError, match="does not exist"):
        load_data_from_filepaths(bogus_file, photoscan_info["texture_abspath"])

def test_convert_between_ras_and_lps():

    # Create example data
    examplePointSet = vtkPolyData()
    points = vtkPoints()

    # Add known points in LPS coordinate space
    points.InsertNextPoint(10, 20, 30)  # Example point in LPS
    points.InsertNextPoint(-15, 25, -35)
    points.InsertNextPoint(0, 0, 0)
    examplePointSet.SetPoints(points)

    expected_ras_points = [
        (-10, -20, 30),  # Flip the sign of X and Y coordinates
        (15, -25, -35),
        (0, 0, 0)
        ]

    convertedPointSet = convert_between_ras_and_lps(examplePointSet)
    converted_points = convertedPointSet.GetPoints()
    for i in range(converted_points.GetNumberOfPoints()):
        # Compare each transformed point with expected RAS points
        converted_point = np.array(convertedPointSet.GetPoint(i))
        expected_point = np.array(expected_ras_points[i])
        np.testing.assert_array_almost_equal(converted_point, expected_point)

def test_convert_numpy_to_vtkimage():

    # Create a dummy 3x2 RGB image as a numpy array with dtype uint16
    dummy_image = np.array(
        [[[65535, 0, 0], [0, 65535, 0]],
         [[0, 0, 65535], [65535, 65535, 65535]],
         [[100, 200, 300], [400, 500, 600]],],
        dtype=np.uint16
    )

    # Convert the numpy array to vtkImageData
    vtk_image = convert_numpy_to_vtkimage(dummy_image)

    # Check dimensions
    assert(vtk_image.GetDimensions() == (dummy_image.shape[1], dummy_image.shape[0],1))

    # Check scalar type and number of components
    vtk_array = vtk_image.GetPointData().GetScalars()
    assert(vtk_array.GetNumberOfComponents() == 3)
    assert(vtk_array.GetDataType() == VTK_UNSIGNED_SHORT)

    # Probe specific coordinates and compare values
    for i in range(dummy_image.shape[0]):
        for j in range(dummy_image.shape[1]):
            # Get the expected RGB values from the numpy array
            expected_rgb = dummy_image[i, j]

            # Compute the flat index of the pixel in vtkImageData
            flat_index = i * dummy_image.shape[1] + j

            # Get the RGB values from vtkImageData
            vtk_rgb = [
                vtk_array.GetComponent(flat_index, 0),
                vtk_array.GetComponent(flat_index, 1),
                vtk_array.GetComponent(flat_index, 2),
            ]

            # Assert that the values are equal
            assert(list(vtk_rgb) ==  list(expected_rgb))

def test_resource_import():
    """Ensure that a meshroom pipeline resource file can be imported"""
    pipeline_path = importlib.resources.files("openlifu.nav.meshroom_pipelines") / "default_pipeline.mg"
    assert pipeline_path.exists()

def test_get_meshroom_pipeline_names():
    """Verify that get_meshroom_pipeline_names gets us a nonempty list of strings"""
    pipeline_names = get_meshroom_pipeline_names()
    assert len(pipeline_names) > 0
    assert isinstance(pipeline_names[0],str)

def test_apply_exif_orientation_numpy():
    """Verify apply_exif_orientation_numpy is consistent with PIL and inverse is consistent"""
    img = np.array([[0,1],[2,3]])

    # Values calculated with PIL ImageOps.exif_transpose
    data = {
        0: np.array([[0, 1], [2, 3]]), # < 1 do nothing
        1: np.array([[0, 1], [2, 3]]),
        2: np.array([[1, 0], [3, 2]]),
        3: np.array([[3, 2], [1, 0]]),
        4: np.array([[2, 3], [0, 1]]),
        5: np.array([[0, 2], [1, 3]]),
        6: np.array([[2, 0], [3, 1]]),
        7: np.array([[3, 1], [2, 0]]),
        8: np.array([[1, 3], [0, 2]]),
        9: np.array([[0, 1], [2, 3]]), # > 8 do nothing
    }

    for orientation, datum in data.items():
        img_tform = apply_exif_orientation_numpy(img, orientation)
        np.testing.assert_array_equal(datum, img_tform)

    # Test inverse is consistent
    for orientation in range(9):
        img_tform = apply_exif_orientation_numpy(img, orientation)
        img_recon = apply_exif_orientation_numpy(img_tform, orientation, inverse=True)
        np.testing.assert_array_equal(img, img_recon)

def test_preprocess_image_modnet():
    """Verify preprocess_image_modnet resize image correctly"""
    assert preprocess_image_modnet(np.zeros((400, 600, 3))).shape == (1,3,384, 576)
    assert preprocess_image_modnet(np.zeros((600, 700, 3))).shape == (1,3,512,576)
    assert preprocess_image_modnet(np.zeros((400, 300, 3))).shape == (1, 3, 672, 512)

def test_make_pairs_sequential_loop():
    """Verify _make_pairs_sequential_loop works correctly"""
    expected_all_way_three = [['a', 'b', 'c'], ['b', 'c']]
    #no window_radius
    assert _make_pairs_sequential_loop(['a','b','c'], window_radius=None) == expected_all_way_three
    #2*window_radius + 1 == list length
    assert _make_pairs_sequential_loop(['a','b','c'], window_radius=1) == expected_all_way_three
    #2*window_radius + 1 > list length
    assert _make_pairs_sequential_loop(['a','b','c'], window_radius=2) == expected_all_way_three
    #usual case
    expected = [['a', 'b', 'c', 'e', 'f'],
                ['b', 'c', 'd', 'f'],
                ['c', 'd', 'e'],
                ['d', 'e', 'f'],
                ['e', 'f']]
    assert _make_pairs_sequential_loop(['a','b','c','d','e','f'], window_radius=2) == expected

def test_make_pairs_sequential():
    """Verify _make_pairs_sequential works correctly"""
    rows = _make_pairs_sequential(['a', 'b', 'c', 'd'],window_radius=1)
    expected = [['a', 'b'],
                ['b', 'c'],
                ['c', 'd']]
    assert rows == expected
    rows = _make_pairs_sequential(['a', 'b', 'c', 'd'],window_radius=2)
    expected = [['a', 'b', 'c'],
                ['b', 'c', 'd'],
                ['c', 'd']]
    assert rows == expected
    expected_exhaustive = [['a', 'b', 'c', 'd'],
                           ['b', 'c', 'd'],
                           ['c', 'd']]
    rows = _make_pairs_sequential(['a', 'b', 'c', 'd'],window_radius=3)
    assert rows == expected_exhaustive
    rows = _make_pairs_sequential(['a', 'b', 'c', 'd'],window_radius=5)
    assert rows == expected_exhaustive

def test_make_pairs_spatial():
    """Verify _make_pairs_spatial works correctly"""
    rows = _make_pairs_spatial(['a', 'b', 'c', 'd', 'e'],num_neighbors=1, locations=[(20,0,0),(10,0,0),(5,0,0),(2,0,0), (0,0,0)])
    expected = [['a', 'b'],
                ['b', 'c'],
                ['c', 'd'],
                ['d', 'e']]
    assert rows == expected
    rows = _make_pairs_spatial(['a', 'b', 'c', 'd', 'e'],num_neighbors=2, locations=[(20,0,0),(10,0,0),(5,0,0),(2,0,0), (0,0,0)])
    expected = [['a', 'b', 'c'],
                ['b', 'c', 'd'],
                ['c', 'd', 'e'],
                ['d', 'e']]
    assert rows == expected
    rows = _make_pairs_spatial(['a', 'b', 'c', 'd', 'e'],num_neighbors=3, locations=[(20,0,0),(10,0,0),(5,0,0),(2,0,0), (0,0,0)])
    expected = [['a', 'b', 'c', 'd'],
                ['b', 'c', 'd', 'e'],
                ['c', 'd', 'e'],
                ['d', 'e']]
    assert rows == expected
    rows = _make_pairs_spatial(['a', 'b', 'c', 'd', 'e'],num_neighbors=5, locations=[(20,0,0),(10,0,0),(5,0,0),(2,0,0), (0,0,0)])
    expected = [['a', 'b', 'c', 'd', 'e'],
                ['b', 'c', 'd', 'e'],
                ['c', 'd', 'e'],
                ['d', 'e']]
    assert rows == expected

    #test on circle
    N = 8
    angles = np.linspace(0, 2 * np.pi, N, endpoint=False)
    x = np.cos(angles)
    y = np.sin(angles)
    z = np.zeros(N)
    locations = list(zip(x, y, z))
    rows = _make_pairs_spatial(['a', 'b', 'c', 'd', 'e', 'f','g', 'h'],num_neighbors=2, locations=locations)
    expected = [['a', 'b', 'h'],
                ['b', 'c'],
                ['c', 'd'],
                ['d', 'e'],
                ['e', 'f'],
                ['f', 'g'],
                ['g', 'h']]
    assert rows == expected
